import torch.nn as nn

class MLP(nn.Module):
    r"""
   Implementation of MLP module.

   Parameters
   ----------
   in_units : int
       Size of input features
   out_units: int
       Size of output units
   num_layers : int
       number of layers in mlp (Default: 1)
   dropout_rate: float
       dropout rate, dropout is applied before every layer (Default: 0.)
   hidden_units: int or list[int]
       size of hidden units or list of hidden unit sizes for each hidden layer (Default: None).
       By default set to the in_units
   bias: bool
        whether to use bias in each layer (Default: true)
   """

    def __init__(self,
                 in_units,
                 out_units,
                 num_layers=1,
                 dropout_rate=0,
                 hidden_units=None,
                 bias=True):

        super(MLP, self).__init__()

        self.dropout = nn.Dropout(dropout_rate)
        self.layers = nn.ModuleList()
        units = [in_units]
        if hidden_units:
            assert isinstance(hidden_units, list) or isinstance(hidden_units, int), "Unsupported type for hidden_units"
        else:
            hidden_units = in_units

        if isinstance(hidden_units, int):
            hidden_units = [hidden_units] * (num_layers-1)
        else:
            assert len(hidden_units) == num_layers - 1, "Mismatch between hidden units per layer and number of layers"

        units += hidden_units
        units.append(out_units)

        for in_unit, out_unit in zip(units[:-1], units[1:]):
            self.layers.append(nn.Linear(in_unit, out_unit, bias=bias))


    def forward(self, x):
        """Forward function.
        Apply mlp on input x``.
        Parameters
        ----------
        x : th.Tensor
            input. Shape: (B, D)
        Returns
        -------
        th.Tensor
            Representation for edge between u and v. Shape: (B, out)
        """
        for layer in self.layers:
            x = self.dropout(x)
            x = layer(x)

        return x




import os
import json
import pickle
import numpy as np
import torch as th
import dgl
from collections.abc import Mapping
from ..utils import get_available_devices

TUNING_JOB_DETAIILS_FILE = 'tuning_job_details.json'
BEST_TRAINING_JOB_KEY = 'BestTrainingJob'
TRAINING_HYPERPARAMETERS_KEY = 'TrainingJobParameters'
TRAINING_JOB_DETAIILS_FILE = 'training_job_details.json'

def get_transform_config():
    r"""

    Returns
    -------
    tuple(data_path: str, model_path: str, devices: list[int], hyperparameters: dict[str, str])
    """
    data_path = os.environ['SM_DATA_DIR']
    output_path = os.environ['SM_OUTPUT_DIR']
    model_path = os.environ['SM_MODEL_DIR']
    devices = get_available_devices()

    if TUNING_JOB_DETAIILS_FILE in os.listdir(output_path):
        tuning_job_details = os.path.join(output_path, TUNING_JOB_DETAIILS_FILE)
        with open(tuning_job_details, 'r') as f:
            params = json.load(f)[BEST_TRAINING_JOB_KEY][TRAINING_HYPERPARAMETERS_KEY]

    else:
        training_job_details = os.path.join(data_path, TRAINING_JOB_DETAIILS_FILE)
        with open(training_job_details, 'r') as f:
            params = json.load(f)[TRAINING_HYPERPARAMETERS_KEY]

    return data_path, model_path, devices, params

def save_node_prediction_model_artifacts(model_path, predictions, graphloader, params, embeddings=None):
    r"""

    save model artifacts necessary for inference for Neptune ML node prediction task

    Parameters
    ----------
    model_path : str
        model directory to save artifacts
    predictions : th.Tensor
        model predictions for all nodes
    graphloader : neptuneml.GraphLoader
        the GraphLoader object
    params : dict[str: str]
        hyperparameter dictionary for the task
    embeddings : th.Tensor (optional)
        computed node embeddings. optional for node prediction tasks
    """
    if th.is_tensor(predictions):
        predictions= predictions.detach().cpu().numpy()
    np.savez(os.path.join(model_path, 'result.npz'), infer_scores=predictions)
    if embeddings is not None:
        if isinstance(embeddings, Mapping):
            save_entity_embeddings(embeddings, model_path, target_ntype=params['target_ntype'])
        else:
            save_entity_embeddings(embeddings, model_path)
    _save_node_prediction_model_metadata(graphloader, model_path, params['target_ntype'])
    _save_inference_config(model_path, params)

def _save_node_prediction_model_metadata(graph_data, path, target_ntype):
    with open(os.path.join(path, 'mapping.info'), "wb") as f:
        node2id = graph_data.node2id
        label_map = graph_data.label_map
        pickle.dump({'node2id': node2id,
                     'label_map': label_map,
                     'target_ntype': target_ntype,
                     'target_ntype_property': graph_data.label_properties[target_ntype]}, f)

def save_link_prediction_model_artifacts(g, model_path, graphloader, params, entity_embeddings, rel_type_embeddings):
    r"""

    save model artifacts necessary for inference for Neptune ML node prediction task

    Parameters
    ----------
    g: dgl.DGLGraph
        the dgl graph
    model_path : str
        model directory to save artifacts
    graphloader : neptuneml.GraphLoader
        the GraphLoader object
    params : dict[str: str]
        hyperparameter dictionary for the task
    entity_embeddings : dict[ntype:str: th.Tensor]
        computed node embeddings.
    rel_type_embeddings : th.Tensor
        learned relation embeddings.
    """
    _save_link_prediction_model_metadata(g, graphloader, entity_embeddings, rel_type_embeddings, params, model_path)
    _save_inference_config(model_path, params)

def _save_link_prediction_model_metadata(g, graph_data, entity_embeddings, rel_type_embeddings, params, path):
    non_reverse_edges = [etype for etype in g.canonical_etypes if 'rev-' not in etype[1]]
    non_reverse_etype_ids = [etid for etid, etype in enumerate(g.canonical_etypes) if etype in non_reverse_edges]
    non_reverse_etype_embeddings = rel_type_embeddings[non_reverse_etype_ids, :]
    if th.is_tensor(non_reverse_etype_embeddings):
        non_reverse_etype_embeddings = non_reverse_etype_embeddings.detach().cpu().numpy()

    eg = dgl.edge_type_subgraph(g, non_reverse_edges)
    et2id_map = {e_type: etid for etid, e_type in enumerate(eg.canonical_etypes)}
    nt2id_map = {n_type: ntid for ntid, n_type in enumerate(eg.ntypes)}

    g = dgl.to_homogeneous(eg)
    localid2globalid = {ntype: th.nonzero(g.ndata[dgl.NTYPE] == ntid).squeeze().numpy()
                        for ntype, ntid in nt2id_map.items()}

    with open(os.path.join(path, 'mapping.info'), "wb") as f:
        node2id = graph_data.node2id
        label_map = graph_data.label_map
        pickle.dump({'model_type': params['model'],
                     'node2id': node2id,
                     'label_map': label_map,
                     'relation2id_map': et2id_map,
                     'nodetype2id_map': nt2id_map,
                     'node2gid': localid2globalid}, f)

    dgl.data.save_graphs(os.path.join(path, 'graph.bin'), g)
    np.save(os.path.join(path, 'relation.npy'), non_reverse_etype_embeddings)
    save_entity_embeddings(entity_embeddings, path, tnid2gnid=localid2globalid)

def save_entity_embeddings(entity_embeddings, path, target_ntype=None, tnid2gnid=None):
    r"""

    save entity embeddings for model inference

    Parameters
    ----------
    entity_embeddings : th.Tensor or dict[str: th.Tensor]
        entity embeddings for a single node type or per node type
    path : str
        model directory to save embeddings
    target_ntype : str (optional)
        the node type of the target.
        if entity_embeddings is a dict causes embedding tensor corresponding to this ntype only to be saved
    tnid2gnid : dict[ntype:str,  np.array[gnid:int]] (optional)
        mapping for node type specific node index to global node index.
        required if entity_embeddings is a dictionary and target_ntype is not set
        since all the embeddings for each node type are concatenated into a single embedding tensor for all nodes
    """
    filename = os.path.join(path, 'entity.npy')
    if isinstance(entity_embeddings, Mapping):
        if target_ntype:
            emb = entity_embeddings[target_ntype]
            if th.is_tensor(emb):
                emb = emb.detach().cpu().numpy()
            np.save(filename, emb)
        else:
            assert tnid2gnid is not None, "Must pass mapping node type node ids to global node ids"
            embed_dim = [embed.shape[1] for embed in entity_embeddings.values()][0]
            total_nodes = sum([embed.shape[0] for embed in entity_embeddings.values()])
            embeddings = np.zeros((total_nodes, embed_dim))
            for ntype, embed in entity_embeddings.items():
                if th.is_tensor(embed):
                    embed = embed.detach().cpu().numpy()
                embeddings[tnid2gnid[ntype], :] = embed
            np.save(filename, embeddings)
    else:
        emb = entity_embeddings.detach().cpu().numpy() if th.is_tensor(entity_embeddings) else entity_embeddings
        np.save(filename, emb)

def _save_inference_config(model_path, params):
    ## generate the inference_config dir and save inference config
    inference_config = {"task": params['task'], "model": params['model'], "name": params['name'],
                        "infer_type": "online-transductive-infer"}
    if 'property' in params:
        inference_config['property'] = params['property']
    inference_config_path = os.path.join(model_path, 'config.json')
    if os.path.exists(inference_config_path):
        with open(inference_config_path, 'r') as f:
            inference_config.update(json.load(f))
    with open(inference_config_path, 'w') as f:
        json.dump(inference_config, f)

def get_test_nids(g, target_ntype=None):
    r"""

    get test node ids from the graph

    Parameters
    ----------
    g : dgl.DGLGraph
        the Graph
    target_ntype : str
        the node type of the targets
   Returns
   -------
   test_idx: th.Tensor
   """
    if 'test_mask' not in g.ndata:
        return None
    test_mask = g.ndata['test_mask'].bool() if target_ntype is None else g.nodes[target_ntype].data['test_mask']
    test_idx = th.arange(test_mask.shape[0])
    test_idx = test_idx[test_mask.bool()]
    return test_idx

def get_test_eids(g, target_etype=None, include_reverse_etypes=False):
    r"""

    get validation node ids from the graph

    Parameters
    ----------
    g : dgl.DGLGraph
        the Graph
    target_etype : str
        the node type of the targets
    include_reverse_etypes : bool (default: False)
        whether to include edge ids from reversed edge types.
    Returns
    -------
    test_idx: th.Tensor or dict[etype:str : th.Tensor]
    """
    test_mask = {target_etype: g.edges[target_etype].data['test_mask']} if target_etype  else g.edata['test_mask']
    if include_reverse_etypes and isinstance(train_mask, Mapping):
        for etype in test_mask:
            src_type, rel_type, dst_type = etype
            if "rev-" in rel_type and (dst_type, rel_type.replace("rev-", ""), src_type) in test_mask:
                test_mask[etype] = g.edata['rev_test_mask'][etype]
    if isinstance(test_mask, Mapping):
        test_idx = {}
        for etype in test_mask:
            idx = th.arange(test_mask[etype].shape[0])
            test_idx[etype] = idx[test_mask[etype].bool()]
    else:
        test_idx = th.arange(test_mask.shape[0])
        test_idx = test_idx[test_mask.bool()]
    return test_idx


def normalize_hyperparameter_keys(params):
    r"""

    format the hyperparameter keys so that the names can be

    Parameters
    ----------
    params : dict[str: str]
        the hyperparameters
   Returns
   -------
   dict[str: str]
   """
    new_params = {}
    for key, value in params.items():
        if "-" in key:
            new_key = key.replace("-", "_")
            new_params[new_key] = value
        else:
            new_params[key] = value

    return new_params




import torch as th
import torch.nn as nn

class DistmultDecoder(nn.Module):
    r"""
    Implementation of Distmult Decoder.

    Paper link: https://arxiv.org/abs/1412.6575

    Suitable when predicting a logit for each edge where
    edge is represented by a pair of arrays of node features
    and optionally an edge type id

    .. math::

        f(u,r,v) = \textbf{u}^{T} \textbf{W}_r \textbf{v} = \sum_{i=1}^{d}\textbf{u}_i \cdot
        diag(\textbf{W}_r)_i \cdot \textbf{v}_i

    Parameters
    ----------
    in_units : int
        Size of input node features
    num_rels: int
        Number of distinct edge types (Default: 1)
    """
    def __init__(self, in_units, num_rels=1):
        super(DistmultDecoder, self).__init__()
        self.w_relation = nn.Parameter(th.Tensor(num_rels, in_units))
        self.num_rels = num_rels
        nn.init.xavier_uniform_(self.w_relation)

    def forward(self, u, v, etype_ids=None):
        """Forward function.
        Compute logits for each pair ``(ufeat[i], vfeat[i])``.
        Parameters
        ----------
        u : th.Tensor
            Source nodes embeddings. Shape: (B, D)
        v : th.Tensor
            Destination nodes embeddings. Shape: (B, D)
        etype_ids: th.Tensor
            Edge type ids for each edge represented by u[i], v[i]
        Returns
        -------
        th.Tensor
            Representation for edge between u and v. Shape: (B, out)
        """
        assert u.shape == v.shape, "Shape mismatch between u and v"
        if etype_ids is None:
            assert self.num_rels == 1, "You must pass in etype_ids if num_rels is more than 1"
        r_emb = self.w_relation[etype_ids] if etype_ids is not None else self.w_relation[0]
        score = th.sum(u * r_emb * v, dim=-1)
        return score

class GraphDistmultDecoder(nn.Module):
    r"""
    Implementation of Distmult Decoder using DGL message passing API.

    Paper link: https://arxiv.org/abs/1412.6575

    Suitable when predicting a logit for each edge using the graph structure

    .. math::

        f(u,r,v) = \textbf{u}^{T} \textbf{W}_r \textbf{v} = \sum_{i=1}^{d}\textbf{u}_i \cdot
        diag(\textbf{W}_r)_i \cdot \textbf{v}_i

    Parameters
    ----------
    in_units : int
        Size of input node features
    num_rels: int
        Number of distinct edge types (Default: 1)
    """
    def __init__(self, in_units, num_rels=1):
        super(GraphDistmultDecoder, self).__init__()
        self.w_relation = nn.Parameter(th.Tensor(num_rels, in_units))
        self.num_rels = num_rels
        nn.init.xavier_uniform_(self.w_relation)

    def forward(self, graph, h):
        """Forward function.
        Compute output for each edge ``(u, v)`` in the graph using `h_u` and h_v`.
        Parameters
        ----------
        graph : dgl.Graph
            The graph structure with the same number of nodes as h
        h : th.Tensor or dict[ntype: th.Tensor]
            Nodes embeddings. Shape: (N, D)
        Returns
        -------
        th.Tensor
            Representation for edge between u and v. Shape: (B, out)
        """
        with graph.local_scope():
            graph.ndata['h'] = h
            for etype in graph.canonical_etypes:
                if graph.num_edges(etype):
                    graph.apply_edges(self.distmult_udf(graph.get_etype_id(etype)), etype=etype)
            return graph.edata['out']

    def distmult_udf(self, etype_id):
        def udf(edges):
            return {'out': th.sum(edges.src['h'] * self.w_relation[etype_id] * edges.dst['h'], dim=-1)}
        return udf


import os
import json
import sklearn
import tqdm
import torch
from collections.abc import Mapping

def classification_report(labels, predictions, return_dict=True, label_map=None, label_indices=None, target_names=None):
    r"""

    Print classification report and return as a dict

    Parameters
    ----------
    labels : th.Tensor
        1 d torch tensor with true class labels. entries must be [0, n_classes-1]
    predictions : th.Tensor
        1 d torch tensor with same shape. entries must be [0, n_classes-1]
    return_dict : boolean default=True
       whether to also return output as dict
    label_map : dict[int, str] default=None
        mapping from label indices to label class names. overrides label_indices and target_names
    label_indices: list[int] default=None
        Optional list of label indices to include in the report.
    target_names: list[str] default=None
        Optional display names matching the labels indices (same order).
    Returns
    -------
    report: dict
    """

    if label_map is not None:
        label_indices, target_names = zip(sorted(label_map))
    l =  labels.cpu().numpy() if torch.is_tensor(labels) else labels
    p = predictions.detach().cpu().numpy() if torch.is_tensor(predictions) else predictions
    print(sklearn.metrics.classification_report(l, p, zero_division=0, labels=label_indices, target_names=target_names))
    if return_dict:
        return sklearn.metrics.classification_report(l, p, output_dict=True, zero_division=0, labels=label_indices,
                                                     target_names=target_names)


def save_eval_metrics(metrics, path):
    r"""

    Save evaluation metrics

    Parameters
    ----------
    metric : dict
        dict container with metric info
    path : str
        directory path to save to
    """
    with open(os.path.join(path, "eval_metrics_info.json"), "w") as f:
        json.dump(metrics, f)


def mrr(node_embeddings, query_edges, scoring_fn, masked_edges=None, src_type=None, dst_type=None, hits=[1]):
    r"""

    Compute and return MRR (mean reciprocal rank) metric for query edges using node embeddings and scoring_fn

    See https://en.wikipedia.org/wiki/Mean_reciprocal_rank for details.


    Parameters
    ----------
    node_embeddings : dict[ntype:str, embed:th.Tensor] or th.Tensor
        node embeddings per node type to be used for computing edge scores
        if th.Tensor and not dictionary then graph must have one node type
    query_edges: tuple(u:th.Tensor, v:th.Tensor, etype_ids:th.Tensor)
        the edges to compute to perturb for computing the mrr in the format (u, v, etype_ids)
        where u is a Tensor of src_node ids, v is a Tensor of dst_node ids and etype_ids is a tensor of edge type ids
    scoring_fn : Callable
        a function that takes embed_u, embed_v and etype_ids and return the score
    masked_edges: tuple(u:th.Tensor, v:th.Tensor, etyped_id:th.Tensor) default=None
        Optional list of edges to exclude when ranking hypothetical edges against target query edge.
    src_type: str default=None
        Optional node type of src nodes in query_edges. Required if node_embeddings is a mapping
    dst_type: str default=None
        Optional node type of dst nodes in query_edges. Required if node_embeddings is a mapping
    hits: list[int] default=[1]
        list of k to return Hits@k
    """
    if isinstance(node_embeddings, Mapping):
        assert src_type is not None and dst_type is not None, "src_type and dst_type must be passed if node_embeddings is a dictionary"
    with torch.no_grad():
        if masked_edges is None:
            triplets_to_filter = []
        else:
            triplets_to_filter = torch.cat([t.reshape(-1, 1) for t in masked_edges], dim=1).tolist()
            triplets_to_filter = {tuple(triplet) for triplet in triplets_to_filter}

        print('Perturbing src nodes...')
        # iterate each query edge and get the rank for all potential src nodes
        ranks_s = _perturb_src_and_get_filtered_rank(node_embeddings, scoring_fn, query_edges, src_type, dst_type, triplets_to_filter)
        print('Perturbing dst nodes...')
        # iterate each query edge and get the rank for all potential dst nodes
        ranks_d = _perturb_dst_and_get_filtered_rank(node_embeddings, scoring_fn, query_edges, src_type, dst_type, triplets_to_filter)

        ranks = torch.cat([ranks_s, ranks_d])
        ranks += 1  # change to 1-indexed

        mrr = torch.mean(1.0 / ranks.float())
        hits_at = []
        for hit in hits:
            avg_count = torch.mean((ranks <= hit).float())
            hits_at.append(avg_count.item())
    return mrr.item(), hits_at

def _perturb_src_and_get_filtered_rank(embedding, scoring_fn, query_edges, src_type=None, dst_type=None, triplets_to_filter=[]):
    num_entities = embedding[src_type].shape[0] if src_type is not None else embedding.shape[0]
    ranks = []
    for target_src, target_dst, target_rel in tqdm.tqdm(zip(*query_edges)):
        target_src, target_dst, target_rel = int(target_src), int(target_dst), int(target_rel)
        src_queries = _get_perturbed_src_queries(triplets_to_filter, target_src, target_rel, target_dst, num_entities)
        emb_s = embedding[src_type][src_queries] if src_type is not None else embedding[src_queries]
        emb_d = embedding[dst_type][target_dst] if dst_type is not None else embedding[target_dst]
        emb_d = emb_d.repeat(emb_s.shape[0], 1)
        rel_ids = torch.full((emb_d.shape[0],), target_rel, dtype=torch.long)
        scores = scoring_fn(emb_s, emb_d, rel_ids)
        _, indices = torch.sort(scores, descending=True)
        target_src_idx = int((src_queries == target_src).nonzero()) # index of current target src
        rank = int((indices == target_src_idx).nonzero()) # rank by score of current target
        ranks.append(rank)
    return torch.LongTensor(ranks)

def _perturb_dst_and_get_filtered_rank(embedding, scoring_fn, query_edges, src_type=None, dst_type=None, triplets_to_filter=[]):
    num_entities = embedding[dst_type].shape[0] if dst_type is not None else embedding.shape[0]
    ranks = []
    for target_src, target_dst, target_rel in tqdm.tqdm(zip(*query_edges)):
        target_src, target_dst, target_rel = int(target_src), int(target_dst), int(target_rel)
        dst_queries = _get_perturbed_dst_queries(triplets_to_filter, target_src, target_rel, target_dst, num_entities)
        emb_d = embedding[dst_type][dst_queries] if dst_type is not None else embedding[dst_queries]
        emb_s = embedding[src_type][target_src] if src_type is not None else embedding[target_src]
        emb_s = emb_s.repeat(emb_d.shape[0], 1)
        rel_ids = torch.full((emb_d.shape[0],), target_rel, dtype=torch.long)
        scores = scoring_fn(emb_s, emb_d, rel_ids)
        _, indices = torch.sort(scores, descending=True)
        target_dst_idx = int((dst_queries == target_dst).nonzero())  # index of current target dst
        rank = int((indices == target_dst_idx).nonzero())  # rank by score of current target
        ranks.append(rank)
    return torch.LongTensor(ranks)

def _get_perturbed_src_queries(triplets_to_filter, target_src, target_rel, target_dst, num_entities):
    return torch.LongTensor([s for s in range(num_entities)
                             if s == target_src or (s, target_dst, target_rel) not in triplets_to_filter])


def _get_perturbed_dst_queries(triplets_to_filter, target_src, target_rel, target_dst, num_entities):
    return torch.LongTensor([d for d in range(num_entities)
                             if d == target_dst or (target_src, d, target_rel) not in triplets_to_filter])


import boto3
import time
import sys
import json
from .cloudwatch_utils import get_log_stream_name
import sagemaker

try:
    boto_session = boto3.DEFAULT_SESSION or boto3.Session()
    region_name = boto_session.region_name
    sm_session = sagemaker.session.Session()
    sagemaker_client = boto3.client(
        service_name='sagemaker',
        region_name=region_name)

    sagemaker_runtime = boto3.client('sagemaker-runtime', region_name=region_name)
except:
    print("Could not initialize boto client")

log_group_name = "/aws/sagemaker/ProcessingJobs"
log_search_url = "https://{region_name}.console.aws.amazon.com/cloudwatch/home?region={region_name}#logStream:group={log_group_name};prefix={job_name}"

_STATUS_CODE_TABLE = {
    "COMPLETED": "Completed",
    "INPROGRESS": "InProgress",
    "FAILED": "Failed",
    "STOPPED": "Stopped",
    "STOPPING": "Stopping",
    "STARTING": "Starting",
    "CREATING": "Creating",
    "INSERVICE": "InService"
}

def wait_for_endpoint(endpoint_name):
    spin = ['-', '/', '|', '\\', '-', '/', '|', '\\']
    spinner = 0
    start_time = 0
    wait = True

    while wait:
        time.sleep(1)
        describe_endpoint_response = sagemaker_client.describe_endpoint(EndpointName=endpoint_name)
        status = describe_endpoint_response['EndpointStatus']
        if status == _STATUS_CODE_TABLE['INSERVICE']:
            print('\rEndpoint [%s] is %-9s... %s' % (endpoint_name, status, "!"),  flush=True, end='')
            break
        elif status == _STATUS_CODE_TABLE['FAILED']:
            print('%s' % ('=' * 80))
            print('Endpoint [%s] creation %s' % (endpoint_name, status))
            print('Failure Reason: %s' % (describe_endpoint_response['FailureReason']))
            break
        else:
            print('\rEndpoint [%s] is %-9s... %s' % (endpoint_name, status, spin[spinner % len(spin)]), flush=True, end='')
            spinner += 1

def wait_for_job(processing_job_name):
    spin = ['-', '/', '|', '\\', '-', '/', '|', '\\']
    spinner = 0
    running = False
    start_time = 0
    wait = True
    log_stream_name = None

    while wait:
        time.sleep(1)
        describe_jobs_response = sagemaker_client.describe_processing_job(ProcessingJobName=processing_job_name)
        status = describe_jobs_response['ProcessingJobStatus']
        if status == _STATUS_CODE_TABLE['COMPLETED'] or status == _STATUS_CODE_TABLE['STOPPED']:
            print('%s' % ('=' * 80))
            print('Job [%s] %s' % (processing_job_name, status))
            break
        elif status == _STATUS_CODE_TABLE['FAILED']:
            print('%s' % ('=' * 80))
            print('Job [%s] %s' % (processing_job_name, status))
            print('Failure Reason: %s' % (describe_jobs_response['FailureReason']))
            print(log_search_url.format(region_name=region_name, log_group_name=log_group_name, job_name=processing_job_name))
            break
        elif status == _STATUS_CODE_TABLE['INPROGRESS']:
            if not log_stream_name:
                print('\rJob [%s] is %-9s... %s' % (processing_job_name, status, spin[spinner % len(spin)]), flush=True, end='')
                spinner += 1
                log_stream_name = get_log_stream_name(log_group_name, processing_job_name)
            if not running and log_stream_name:
                running = True
                print('\rJob [%s] is %-9s...' % (processing_job_name, status))
                print('Output [%s]:\n%s' % (log_stream_name, '=' * 80))
            if log_stream_name:
                sm_session.logs_for_processing_job(processing_job_name, wait=True)
        else:
            print('\rJob [%s] is %-9s... %s' % (processing_job_name, status, spin[spinner % len(spin)]), flush=True, end='')
            spinner += 1

def parse_arn_for_job_name(processing_job_arn):
    return processing_job_arn.split("/")[-1]

def get_processing_job_details(job_name):
    return sagemaker_client.describe_processing_job(ProcessingJobName=job_name)

def get_hyperparameter_tuning_job_details(job_name):
    return sagemaker_client.describe_hyper_parameter_tuning_job(HyperParameterTuningJobName=job_name)

def invoke_endpoint(endpoint_name, data, content_type='application/json', response_format='application/json'):
    payload = json.dumps(data)
    response = sagemaker_runtime.invoke_endpoint(EndpointName=endpoint_name,
                                                 ContentType=content_type,
                                                 Accept=response_format,
                                                 Body=payload)
    return response['Body'].read().decode()


import boto3

try:
    boto_session = boto3.DEFAULT_SESSION or boto3.Session()
    region_name = boto_session.region_name
    neptune_client = boto3.client(service_name='neptune', region_name=region_name)
except:
    print("Could not initialize boto client")

def get_db_cluster_parameter_group(neptune_host):
    identifier = neptune_host.split(".")[0]
    describe_cluster_response = neptune_client.describe_db_clusters(DBClusterIdentifier=identifier)
    cluster_parameter_group = describe_cluster_response["DBClusters"][0]["DBClusterParameterGroup"]
    return cluster_parameter_group

def get_neptune_ml_iam_role_parameter_group(neptune_host):
    cluster_parameter_group_name = get_db_cluster_parameter_group(neptune_host)
    describe_parameter_group_response = neptune_client.describe_db_cluster_parameters(DBClusterParameterGroupName=cluster_parameter_group_name)
    if "Parameters" in describe_parameter_group_response:
        for parameter in describe_parameter_group_response["Parameters"]:
            if parameter["ParameterName"] == "neptune_ml_iam_role":
                if parameter["ParameterValue"]:
                    return parameter["ParameterValue"]
                else:
                    print("neptune_ml_iam_role Parameter not assigned")
        print("Could not find neptune_ml_iam_role parameter in cluster parameter group")
    raise


class EdgeRelGraphConv(nn.Module):
    r"""Relational graph convolution layer.

    Relational graph convolution is introduced in "`Modeling Relational Data with Graph
    Convolutional Networks <https://arxiv.org/abs/1703.06103>`__"
    and can be described as below:

    .. math::

       h_i^{(l+1)} = \sigma(\sum_{r\in\mathcal{R}}
       \sum_{j\in\mathcal{N}^r(i)}\frac{1}{c_{i,r}}W_r^{(l)}h_j^{(l)}+W_0^{(l)}h_i^{(l)})

    where :math:`\mathcal{N}^r(i)` is the neighbor set of node :math:`i` w.r.t. relation
    :math:`r`. :math:`c_{i,r}` is the normalizer equal
    to :math:`|\mathcal{N}^r(i)|`. :math:`\sigma` is an activation function. :math:`W_0`
    is the self-loop weight.

    The basis regularization decomposes :math:`W_r` by:

    .. math::

       W_r^{(l)} = \sum_{b=1}^B a_{rb}^{(l)}V_b^{(l)}

    where :math:`B` is the number of bases, :math:`V_b^{(l)}` are linearly combined
    with coefficients :math:`a_{rb}^{(l)}`.

    The block-diagonal-decomposition regularization decomposes :math:`W_r` into :math:`B`
    number of block diagonal matrices. We refer :math:`B` as the number of bases.

    The block regularization decomposes :math:`W_r` by:

    .. math::

       W_r^{(l)} = \oplus_{b=1}^B Q_{rb}^{(l)}

    where :math:`B` is the number of bases, :math:`Q_{rb}^{(l)}` are block
    bases with shape :math:`R^{(d^{(l+1)}/B)*(d^{l}/B)}`.

    Parameters
    ----------
    in_feat : int
        Input feature size; i.e, the number of dimensions of :math:`h_j^{(l)}`.
    out_feat : int
        Output feature size; i.e., the number of dimensions of :math:`h_i^{(l+1)}`.
    num_rels : int
        Number of relations.
    active_rels: list[int]
        Relations that are active for layer
    regularizer : str
        Which weight regularizer to use "basis" or "bdd".
        "basis" is short for basis-diagonal-decomposition.
        "bdd" is short for block-diagonal-decomposition.
    num_bases : int, optional
        Number of bases. If is none, use number of relations. Default: ``None``.
    bias : bool, optional
        True if bias is added. Default: ``True``.
    activation : callable, optional
        Activation function. Default: ``None``.
    self_loop : bool, optional
        True to include self loop message. Default: ``True``.
    low_mem : bool, optional
        True to use low memory implementation of relation message passing function. Default: False.
        This option trades speed with memory consumption, and will slowdown the forward/backward.
        Turn it on when you encounter OOM problem during training or evaluation. Default: ``False``.
    dropout : float, optional
        Dropout rate. Default: ``0.0``
    layer_norm: float, optional
        Add layer norm. Default: ``False``
    edge_dim : int
        The dimension for the edge transformed features. If None then the layer behaves as the original RGCN
        otherwise it will take into account the edge features as dictated by edge_aggregation_type
    edge_aggregation_type :
        Type of aggregation done for the edge transformed features it can anything from "sum", "product", "concat"

    Examples
    --------
    >>> import dgl
    >>> import numpy as np
    >>> import torch as th
    >>> from dgl.nn import RelGraphConv
    >>>
    >>> g = dgl.graph(([0,1,2,3,2,5], [1,2,3,4,0,3]))
    >>> feat = th.ones(6, 10)
    >>> conv = RelGraphConv(10, 2, 3, regularizer='basis', num_bases=2)
    >>> conv.weight.shape
    torch.Size([2, 10, 2])
    >>> etype = th.tensor(np.array([0,1,2,0,1,2]).astype(np.int64))
    >>> res = conv(g, feat, etype)
    >>> res
    tensor([[ 0.3996, -2.3303],
            [-0.4323, -0.1440],
            [ 0.3996, -2.3303],
            [ 2.1046, -2.8654],
            [-0.4323, -0.1440],
            [-0.1309, -1.0000]], grad_fn=<AddBackward0>)

    >>> # One-hot input
    >>> one_hot_feat = th.tensor(np.array([0,1,2,3,4,5]).astype(np.int64))
    >>> res = conv(g, one_hot_feat, etype)
    >>> res
    tensor([[ 0.5925,  0.0985],
            [-0.3953,  0.8408],
            [-0.9819,  0.5284],
            [-1.0085, -0.1721],
            [ 0.5962,  1.2002],
            [ 0.0365, -0.3532]], grad_fn=<AddBackward0>)
    """
    def __init__(self,
                 in_feat,
                 out_feat,
                 num_rels,
                 active_rels=None,
                 regularizer="basis",
                 num_bases=None,
                 bias=True,
                 activation=None,
                 self_loop=True,
                 low_mem=False,
                 dropout=0.0,
                 edge_dim=None,
                 edge_aggregation_type="sum",
                 layer_norm=False):class EdgeRelGraphConv(nn.Module):
    r"""Relational graph convolution layer.

    Relational graph convolution is introduced in "`Modeling Relational Data with Graph
    Convolutional Networks <https://arxiv.org/abs/1703.06103>`__"
    and can be described as below:

    .. math::

       h_i^{(l+1)} = \sigma(\sum_{r\in\mathcal{R}}
       \sum_{j\in\mathcal{N}^r(i)}\frac{1}{c_{i,r}}W_r^{(l)}h_j^{(l)}+W_0^{(l)}h_i^{(l)})

    where :math:`\mathcal{N}^r(i)` is the neighbor set of node :math:`i` w.r.t. relation
    :math:`r`. :math:`c_{i,r}` is the normalizer equal
    to :math:`|\mathcal{N}^r(i)|`. :math:`\sigma` is an activation function. :math:`W_0`
    is the self-loop weight.

    The basis regularization decomposes :math:`W_r` by:

    .. math::

       W_r^{(l)} = \sum_{b=1}^B a_{rb}^{(l)}V_b^{(l)}

    where :math:`B` is the number of bases, :math:`V_b^{(l)}` are linearly combined
    with coefficients :math:`a_{rb}^{(l)}`.

    The block-diagonal-decomposition regularization decomposes :math:`W_r` into :math:`B`
    number of block diagonal matrices. We refer :math:`B` as the number of bases.

    The block regularization decomposes :math:`W_r` by:

    .. math::

       W_r^{(l)} = \oplus_{b=1}^B Q_{rb}^{(l)}

    where :math:`B` is the number of bases, :math:`Q_{rb}^{(l)}` are block
    bases with shape :math:`R^{(d^{(l+1)}/B)*(d^{l}/B)}`.

    Parameters
    ----------
    in_feat : int
        Input feature size; i.e, the number of dimensions of :math:`h_j^{(l)}`.
    out_feat : int
        Output feature size; i.e., the number of dimensions of :math:`h_i^{(l+1)}`.
    num_rels : int
        Number of relations.
    active_rels: list[int]
        Relations that are active for layer
    regularizer : str
        Which weight regularizer to use "basis" or "bdd".
        "basis" is short for basis-diagonal-decomposition.
        "bdd" is short for block-diagonal-decomposition.
    num_bases : int, optional
        Number of bases. If is none, use number of relations. Default: ``None``.
    bias : bool, optional
        True if bias is added. Default: ``True``.
    activation : callable, optional
        Activation function. Default: ``None``.
    self_loop : bool, optional
        True to include self loop message. Default: ``True``.
    low_mem : bool, optional
        True to use low memory implementation of relation message passing function. Default: False.
        This option trades speed with memory consumption, and will slowdown the forward/backward.
        Turn it on when you encounter OOM problem during training or evaluation. Default: ``False``.
    dropout : float, optional
        Dropout rate. Default: ``0.0``
    layer_norm: float, optional
        Add layer norm. Default: ``False``
    edge_dim : int
        The dimension for the edge transformed features. If None then the layer behaves as the original RGCN
        otherwise it will take into account the edge features as dictated by edge_aggregation_type
    edge_aggregation_type :
        Type of aggregation done for the edge transformed features it can anything from "sum", "product", "concat"

    Examples
    --------
    >>> import dgl
    >>> import numpy as np
    >>> import torch as th
    >>> from dgl.nn import RelGraphConv
    >>>
    >>> g = dgl.graph(([0,1,2,3,2,5], [1,2,3,4,0,3]))
    >>> feat = th.ones(6, 10)
    >>> conv = RelGraphConv(10, 2, 3, regularizer='basis', num_bases=2)
    >>> conv.weight.shape
    torch.Size([2, 10, 2])
    >>> etype = th.tensor(np.array([0,1,2,0,1,2]).astype(np.int64))
    >>> res = conv(g, feat, etype)
    >>> res
    tensor([[ 0.3996, -2.3303],
            [-0.4323, -0.1440],
            [ 0.3996, -2.3303],
            [ 2.1046, -2.8654],
            [-0.4323, -0.1440],
            [-0.1309, -1.0000]], grad_fn=<AddBackward0>)

    >>> # One-hot input
    >>> one_hot_feat = th.tensor(np.array([0,1,2,3,4,5]).astype(np.int64))
    >>> res = conv(g, one_hot_feat, etype)
    >>> res
    tensor([[ 0.5925,  0.0985],
            [-0.3953,  0.8408],
            [-0.9819,  0.5284],
            [-1.0085, -0.1721],
            [ 0.5962,  1.2002],
            [ 0.0365, -0.3532]], grad_fn=<AddBackward0>)
    """
    def __init__(self,
                 in_feat,
                 out_feat,
                 num_rels,
                 active_rels=None,
                 regularizer="basis",
                 num_bases=None,
                 bias=True,
                 activation=None,
                 self_loop=True,
                 low_mem=False,
                 dropout=0.0,
                 edge_dim=None,
                 edge_aggregation_type="sum",
                 layer_norm=False):
        super(EdgeRelGraphConv, self).__init__()
        self.in_feat = in_feat
        self.out_feat = out_feat
        self.num_rels = num_rels
        self.regularizer = regularizer
        self.num_bases = num_bases
        self.active_rels = range(num_rels) if active_rels is None else active_rels
        if self.num_bases is None or self.num_bases > self.num_rels or self.num_bases <= 0:
            self.num_bases = self.num_rels
        self.bias = bias
        self.activation = activation
        self.self_loop = self_loop
        self.low_mem = low_mem
        self.layer_norm = layer_norm
        self.edge_dim =edge_dim
        self.edge_aggregation_type = edge_aggregation_type

        if regularizer == "basis":
            # add basis weights
            self.weight = nn.Parameter(th.Tensor(self.num_bases, self.in_feat, self.out_feat))
            if self.num_bases < self.num_rels:
                # linear combination coefficients
                self.w_comp = nn.Parameter(th.Tensor(self.num_rels, self.num_bases))
            nn.init.xavier_uniform_(self.weight, gain=nn.init.calculate_gain('relu'))
            if self.num_bases < self.num_rels:
                nn.init.xavier_uniform_(self.w_comp,
                                        gain=nn.init.calculate_gain('relu'))
            # message func
            self.message_func = self.basis_message_func
        elif regularizer == "bdd":
            if in_feat % self.num_bases != 0 or out_feat % self.num_bases != 0:
                raise ValueError(
                    'Feature size must be a multiplier of num_bases (%d).'
                    % self.num_bases
                )
            # add block diagonal weights
            self.submat_in = in_feat // self.num_bases
            self.submat_out = out_feat // self.num_bases

            # assuming in_feat and out_feat are both divisible by num_bases
            self.weight = nn.Parameter(th.Tensor(
                self.num_rels, self.num_bases * self.submat_in * self.submat_out))
            nn.init.xavier_uniform_(self.weight, gain=nn.init.calculate_gain('relu'))
            # message func
            self.message_func = self.bdd_message_func
        else:
            raise ValueError("Regularizer must be either 'basis' or 'bdd'")

        # bias
        if self.bias:
            self.h_bias = nn.Parameter(th.Tensor(out_feat))
            nn.init.zeros_(self.h_bias)

        # layer norm
        if self.layer_norm:
            self.layer_norm_weight = nn.LayerNorm(out_feat, elementwise_affine=True)

        # weight for self loop
        if self.self_loop:
            self.loop_weight = nn.Parameter(th.Tensor(in_feat, out_feat))
            nn.init.xavier_uniform_(self.loop_weight,
                                    gain=nn.init.calculate_gain('relu'))
        if edge_dim is not None:
            self.edge_feat_transform_weight= nn.Linear(self.edge_dim if self.edge_aggregation_type
                                            != "concat" else self.edge_dim+self.out_feat,self.out_feat)
        self.dropout = nn.Dropout(dropout)

    def edge_feat_transform(self, msg, edge_feats):
        type = self.edge_aggregation_type
        if type== "sum":
            msg = msg + self.edge_feat_transform_weight(edge_feats)
        elif type == "product":
            msg = msg * self.edge_feat_transform_weight(edge_feats)
        elif type == "concat":
            msg =  self.edge_feat_transform_weight(th.cat((edge_feats,msg),dim=1))
        else:
            raise TypeError('Not supported message aggregate  type.')
        return msg

    def basis_message_func(self, edges, etypes):
        """Message function for basis regularizer.

        Parameters
        ----------
        edges : dgl.EdgeBatch
            Input to DGL message UDF.
        etypes : torch.Tensor or list[int]
            Edge type data. Could be either:

                * An :math:`(|E|,)` dense tensor. Each element corresponds to the edge's type ID.
                  Preferred format if ``lowmem == False``.
                * An integer list. The i^th element is the number of edges of the i^th type.
                  This requires the input graph to store edges sorted by their type IDs.
                  Preferred format if ``lowmem == True``.
        """
        if self.num_bases < self.num_rels:
            # generate all weights from bases
            weight = self.weight.view(self.num_bases,
                                      self.in_feat * self.out_feat)
            weight = th.matmul(self.w_comp, weight).view(
                self.num_rels, self.in_feat, self.out_feat)
        else:
            weight = self.weight

        h = edges.src['h']
        device = h.device

        if h.shape[0] == 0:
            return {'msg': th.tensor((), device=device)}

        if h.dtype == th.int64 and h.ndim == 1:
            # Each element is the node's ID. Use index select: weight[etypes, h, :]
            # The following is a faster version of it.
            if isinstance(etypes, list):
                etypes = th.repeat_interleave(th.arange(len(etypes), device=device),
                                              th.tensor(etypes, device=device))
            weight = weight.view(-1, weight.shape[2])
            flatidx = etypes * weight.shape[1] + h
            msg = weight.index_select(0, flatidx)
        elif self.low_mem:
            # A more memory-friendly implementation.
            # Calculate msg @ W_r before put msg into edge.
            assert isinstance(etypes, list)
            h_t = th.split(h, etypes)
            msg = []
            for etype in range(self.num_rels):
                if h_t[etype].shape[0] == 0:
                    continue
                if etype in self.active_rels: # check if this etype is active
                    msg.append(th.zeros(h_t[etype].shape))
                else:
                    msg.append(th.matmul(h_t[etype], weight[etype]))
            msg = th.cat(msg)
        else:
            # Use batched matmult
            active_idx = th.nonzero(edges.data["mask"]).squeeze()
            msg = th.zeros(h.shape[0], self.out_feat, device=device)
            if isinstance(etypes, list):
                etypes = th.repeat_interleave(th.arange(len(etypes), device=device),
                                              th.tensor(etypes, device=device))
            weight = weight.index_select(0, etypes)
            msg[active_idx] = th.bmm(h.unsqueeze(1), weight).squeeze(1)[active_idx]

        if 'norm' in edges.data:
            msg = msg * edges.data['norm']
        if 'feat' in edges.data:
            edge_feat_seeds = th.nonzero(th.logical_and(edges.data["efeat_bool"], edges.data["mask"]).int()).squeeze()
            if edge_feat_seeds.shape[0] != 0:
                msg[edge_feat_seeds] = self.edge_feat_transform(msg[edge_feat_seeds],edges.data["feat"][edge_feat_seeds])
        return {'msg': msg}

    def bdd_message_func(self, edges, etypes):
        """Message function for block-diagonal-decomposition regularizer.

        Parameters
        ----------
        edges : dgl.EdgeBatch
            Input to DGL message UDF.
        etypes : torch.Tensor or list[int]
            Edge type data. Could be either:

                * An :math:`(|E|,)` dense tensor. Each element corresponds to the edge's type ID.
                  Preferred format if ``lowmem == False``.
                * An integer list. The i^th element is the number of edges of the i^th type.
                  This requires the input graph to store edges sorted by their type IDs.
                  Preferred format if ``lowmem == True``.
        """
        h = edges.src['h']
        device = h.device

        if h.dtype == th.int64 and h.ndim == 1:
            raise TypeError('Block decomposition does not allow integer ID feature.')

        if self.low_mem:
            # A more memory-friendly implementation.
            # Calculate msg @ W_r before put msg into edge.
            assert isinstance(etypes, list)
            h_t = th.split(h, etypes)
            msg = []
            for etype in range(self.num_rels):
                if h_t[etype].shape[0] == 0:
                    continue
                if etype in self.active_rels: # check if this etype is active
                    msg.append(th.zeros(h_t[etype].shape))
                else:
                    tmp_w = self.weight[etype].view(self.num_bases, self.submat_in, self.submat_out)
                    tmp_h = h_t[etype].view(-1, self.num_bases, self.submat_in)
                    msg.append(th.einsum('abc,bcd->abd', tmp_h, tmp_w).reshape(-1, self.out_feat))
            msg = th.cat(msg)
        else:
            # Use batched matmult
            active_idx = th.nonzero(edges.data["mask"]).squeeze()
            msg = th.zeros(h.shape[0], self.out_feat, device=device)
            if isinstance(etypes, list):
                etypes = th.repeat_interleave(th.arange(len(etypes), device=device),
                                              th.tensor(etypes, device=device))
            weight = self.weight.index_select(0, etypes).view(
                -1, self.submat_in, self.submat_out)
            node = h.view(-1, 1, self.submat_in)
            msg[active_idx] = th.bmm(node, weight).view(-1, self.out_feat)[active_idx]
        if 'norm' in edges.data:
            msg = msg * edges.data['norm']
        if 'feat' in edges.data:
            edge_feat_seeds = th.nonzero(th.logical_and(edges.data["efeat_bool"], edges.data["mask"]).int()).squeeze()
            if edge_feat_seeds.shape[0] != 0:
                msg[edge_feat_seeds] = self.edge_feat_transform(msg[edge_feat_seeds], edges.data["feat"][edge_feat_seeds])
        return {'msg': msg}

    def forward(self, g, feat, etypes, norm=None, edge_feat = None, efeat_bool=None):
        """Forward computation.

        Parameters
        ----------
        g : DGLGraph
            The graph.
        feat : torch.Tensor
            Input node features. Could be either

                * :math:`(|V|, D)` dense tensor
                * :math:`(|V|,)` int64 vector, representing the categorical values of each
                  node. It then treat the input feature as an one-hot encoding feature.
        etypes : torch.Tensor or list[int]
            Edge type data. Could be either

                * An :math:`(|E|,)` dense tensor. Each element corresponds to the edge's type ID.
                  Preferred format if ``lowmem == False``.
                * An integer list. The i^th element is the number of edges of the i^th type.
                  This requires the input graph to store edges sorted by their type IDs.
                  Preferred format if ``lowmem == True``.
        norm : torch.Tensor
            Optional edge normalizer tensor. Shape: :math:`(|E|, 1)`.
        edge_feat : torch.Tensor
            Optional argument for the edge features
        efeat_bool : torch.Tensor
            Optional argument to specify which edges in the batch are associated with features for the rest the layer
            behaves as an RGCN

        Returns
        -------
        torch.Tensor
            New node features.

        Notes
        -----
        Under the ``low_mem`` mode, DGL will sort the graph based on the edge types
        and compute message passing one type at a time. DGL recommends sorts the
        graph beforehand (and cache it if possible) and provides the integer list
        format to the ``etypes`` argument. Use DGL's :func:`~dgl.to_homogeneous` API
        to get a sorted homogeneous graph from a heterogeneous graph. Pass ``return_count=True``
        to it to get the ``etypes`` in integer list.
        """
        if isinstance(etypes, th.Tensor):
            if len(etypes) != g.num_edges():
                raise DGLError('"etypes" tensor must have length equal to the number of edges'
                               ' in the graph. But got {} and {}.'.format(
                                   len(etypes), g.num_edges()))
            if self.low_mem and not (feat.dtype == th.int64 and feat.ndim == 1):
                # Low-mem optimization is not enabled for node ID input. When enabled,
                # it first sorts the graph based on the edge types (the sorting will not
                # change the node IDs). It then converts the etypes tensor to an integer
                # list, where each element is the number of edges of the type.
                # Sort the graph based on the etypes
                sorted_etypes, index = th.sort(etypes)
                g = dgl.edge_subgraph(g, index, relabel_nodes=False)
                # Create a new etypes to be an integer list of number of edges.
                pos = th.searchsorted(sorted_etypes, th.arange(self.num_rels, device=g.device))
                num = th.tensor([len(etypes)], device=g.device)
                etypes = (th.cat([pos[1:], num]) - pos).tolist()
                if norm is not None:
                    norm = norm[index]

        with g.local_scope():
            g.srcdata['h'] = feat
            if norm is not None:
                g.edata['norm'] = norm
            if edge_feat is not None:
                g.edata['feat'] = edge_feat
                g.edata['efeat_bool'] = efeat_bool
            g.edata['mask'] = sum(etypes==i for i in self.active_rels)
            if self.self_loop:
                loop_message = matmul_maybe_select(feat[:g.number_of_dst_nodes()],
                                                   self.loop_weight)
            # message passing
            g.update_all(functools.partial(self.message_func, etypes=etypes),
                         fn.sum(msg='msg', out='h'))

            # apply bias and activation
            if 'h' in g.dstdata and len(g.dstdata['h'].shape) == 2:
                node_repr = g.dstdata['h']
            else:
                node_repr = th.zeros((g.number_of_dst_nodes(), self.out_feat),
                                     dtype=feat.dtype,
                                     device=feat.device)
            if self.layer_norm:
                node_repr = self.layer_norm_weight(node_repr)
            if self.bias:
                node_repr = node_repr + self.h_bias
            if self.self_loop:
                node_repr = node_repr + loop_message
            if self.activation:
                node_repr = self.activation(node_repr)
            node_repr = self.dropout(node_repr)
            return node_repr
        super(EdgeRelGraphConv, self).__init__()
        self.in_feat = in_feat
        self.out_feat = out_feat
        self.num_rels = num_rels
        self.regularizer = regularizer
        self.num_bases = num_bases
        self.active_rels = range(num_rels) if active_rels is None else active_rels
        if self.num_bases is None or self.num_bases > self.num_rels or self.num_bases <= 0:
            self.num_bases = self.num_rels
        self.bias = bias
        self.activation = activation
        self.self_loop = self_loop
        self.low_mem = low_mem
        self.layer_norm = layer_norm
        self.edge_dim =edge_dim
        self.edge_aggregation_type = edge_aggregation_type

        if regularizer == "basis":
            # add basis weights
            self.weight = nn.Parameter(th.Tensor(self.num_bases, self.in_feat, self.out_feat))
            if self.num_bases < self.num_rels:
                # linear combination coefficients
                self.w_comp = nn.Parameter(th.Tensor(self.num_rels, self.num_bases))
            nn.init.xavier_uniform_(self.weight, gain=nn.init.calculate_gain('relu'))
            if self.num_bases < self.num_rels:
                nn.init.xavier_uniform_(self.w_comp,
                                        gain=nn.init.calculate_gain('relu'))
            # message func
            self.message_func = self.basis_message_func
        elif regularizer == "bdd":
            if in_feat % self.num_bases != 0 or out_feat % self.num_bases != 0:
                raise ValueError(
                    'Feature size must be a multiplier of num_bases (%d).'
                    % self.num_bases
                )
            # add block diagonal weights
            self.submat_in = in_feat // self.num_bases
            self.submat_out = out_feat // self.num_bases

            # assuming in_feat and out_feat are both divisible by num_bases
            self.weight = nn.Parameter(th.Tensor(
                self.num_rels, self.num_bases * self.submat_in * self.submat_out))
            nn.init.xavier_uniform_(self.weight, gain=nn.init.calculate_gain('relu'))
            # message func
            self.message_func = self.bdd_message_func
        else:
            raise ValueError("Regularizer must be either 'basis' or 'bdd'")

        # bias
        if self.bias:
            self.h_bias = nn.Parameter(th.Tensor(out_feat))
            nn.init.zeros_(self.h_bias)

        # layer norm
        if self.layer_norm:
            self.layer_norm_weight = nn.LayerNorm(out_feat, elementwise_affine=True)

        # weight for self loop
        if self.self_loop:
            self.loop_weight = nn.Parameter(th.Tensor(in_feat, out_feat))
            nn.init.xavier_uniform_(self.loop_weight,
                                    gain=nn.init.calculate_gain('relu'))
        if edge_dim is not None:
            self.edge_feat_transform_weight= nn.Linear(self.edge_dim if self.edge_aggregation_type
                                            != "concat" else self.edge_dim+self.out_feat,self.out_feat)
        self.dropout = nn.Dropout(dropout)

    def edge_feat_transform(self, msg, edge_feats):
        type = self.edge_aggregation_type
        if type== "sum":
            msg = msg + self.edge_feat_transform_weight(edge_feats)
        elif type == "product":
            msg = msg * self.edge_feat_transform_weight(edge_feats)
        elif type == "concat":
            msg =  self.edge_feat_transform_weight(th.cat((edge_feats,msg),dim=1))
        else:
            raise TypeError('Not supported message aggregate  type.')
        return msg

    def basis_message_func(self, edges, etypes):
        """Message function for basis regularizer.

        Parameters
        ----------
        edges : dgl.EdgeBatch
            Input to DGL message UDF.
        etypes : torch.Tensor or list[int]
            Edge type data. Could be either:

                * An :math:`(|E|,)` dense tensor. Each element corresponds to the edge's type ID.
                  Preferred format if ``lowmem == False``.
                * An integer list. The i^th element is the number of edges of the i^th type.
                  This requires the input graph to store edges sorted by their type IDs.
                  Preferred format if ``lowmem == True``.
        """
        if self.num_bases < self.num_rels:
            # generate all weights from bases
            weight = self.weight.view(self.num_bases,
                                      self.in_feat * self.out_feat)
            weight = th.matmul(self.w_comp, weight).view(
                self.num_rels, self.in_feat, self.out_feat)
        else:
            weight = self.weight

        h = edges.src['h']
        device = h.device

        if h.shape[0] == 0:
            return {'msg': th.tensor((), device=device)}

        if h.dtype == th.int64 and h.ndim == 1:
            # Each element is the node's ID. Use index select: weight[etypes, h, :]
            # The following is a faster version of it.
            if isinstance(etypes, list):
                etypes = th.repeat_interleave(th.arange(len(etypes), device=device),
                                              th.tensor(etypes, device=device))
            weight = weight.view(-1, weight.shape[2])
            flatidx = etypes * weight.shape[1] + h
            msg = weight.index_select(0, flatidx)
        elif self.low_mem:
            # A more memory-friendly implementation.
            # Calculate msg @ W_r before put msg into edge.
            assert isinstance(etypes, list)
            h_t = th.split(h, etypes)
            msg = []
            for etype in range(self.num_rels):
                if h_t[etype].shape[0] == 0:
                    continue
                if etype in self.active_rels: # check if this etype is active
                    msg.append(th.zeros(h_t[etype].shape))
                else:
                    msg.append(th.matmul(h_t[etype], weight[etype]))
            msg = th.cat(msg)
        else:
            # Use batched matmult
            active_idx = th.nonzero(edges.data["mask"]).squeeze()
            msg = th.zeros(h.shape[0], self.out_feat, device=device)
            if isinstance(etypes, list):
                etypes = th.repeat_interleave(th.arange(len(etypes), device=device),
                                              th.tensor(etypes, device=device))
            weight = weight.index_select(0, etypes)
            msg[active_idx] = th.bmm(h.unsqueeze(1), weight).squeeze(1)[active_idx]

        if 'norm' in edges.data:
            msg = msg * edges.data['norm']
        if 'feat' in edges.data:
            edge_feat_seeds = th.nonzero(th.logical_and(edges.data["efeat_bool"], edges.data["mask"]).int()).squeeze()
            if edge_feat_seeds.shape[0] != 0:
                msg[edge_feat_seeds] = self.edge_feat_transform(msg[edge_feat_seeds],edges.data["feat"][edge_feat_seeds])
        return {'msg': msg}

    def bdd_message_func(self, edges, etypes):
        """Message function for block-diagonal-decomposition regularizer.

        Parameters
        ----------
        edges : dgl.EdgeBatch
            Input to DGL message UDF.
        etypes : torch.Tensor or list[int]
            Edge type data. Could be either:

                * An :math:`(|E|,)` dense tensor. Each element corresponds to the edge's type ID.
                  Preferred format if ``lowmem == False``.
                * An integer list. The i^th element is the number of edges of the i^th type.
                  This requires the input graph to store edges sorted by their type IDs.
                  Preferred format if ``lowmem == True``.
        """
        h = edges.src['h']
        device = h.device

        if h.dtype == th.int64 and h.ndim == 1:
            raise TypeError('Block decomposition does not allow integer ID feature.')

        if self.low_mem:
            # A more memory-friendly implementation.
            # Calculate msg @ W_r before put msg into edge.
            assert isinstance(etypes, list)
            h_t = th.split(h, etypes)
            msg = []
            for etype in range(self.num_rels):
                if h_t[etype].shape[0] == 0:
                    continue
                if etype in self.active_rels: # check if this etype is active
                    msg.append(th.zeros(h_t[etype].shape))
                else:
                    tmp_w = self.weight[etype].view(self.num_bases, self.submat_in, self.submat_out)
                    tmp_h = h_t[etype].view(-1, self.num_bases, self.submat_in)
                    msg.append(th.einsum('abc,bcd->abd', tmp_h, tmp_w).reshape(-1, self.out_feat))
            msg = th.cat(msg)
        else:
            # Use batched matmult
            active_idx = th.nonzero(edges.data["mask"]).squeeze()
            msg = th.zeros(h.shape[0], self.out_feat, device=device)
            if isinstance(etypes, list):
                etypes = th.repeat_interleave(th.arange(len(etypes), device=device),
                                              th.tensor(etypes, device=device))
            weight = self.weight.index_select(0, etypes).view(
                -1, self.submat_in, self.submat_out)
            node = h.view(-1, 1, self.submat_in)
            msg[active_idx] = th.bmm(node, weight).view(-1, self.out_feat)[active_idx]
        if 'norm' in edges.data:
            msg = msg * edges.data['norm']
        if 'feat' in edges.data:
            edge_feat_seeds = th.nonzero(th.logical_and(edges.data["efeat_bool"], edges.data["mask"]).int()).squeeze()
            if edge_feat_seeds.shape[0] != 0:
                msg[edge_feat_seeds] = self.edge_feat_transform(msg[edge_feat_seeds], edges.data["feat"][edge_feat_seeds])
        return {'msg': msg}

    def forward(self, g, feat, etypes, norm=None, edge_feat = None, efeat_bool=None):
        """Forward computation.

        Parameters
        ----------
        g : DGLGraph
            The graph.
        feat : torch.Tensor
            Input node features. Could be either

                * :math:`(|V|, D)` dense tensor
                * :math:`(|V|,)` int64 vector, representing the categorical values of each
                  node. It then treat the input feature as an one-hot encoding feature.
        etypes : torch.Tensor or list[int]
            Edge type data. Could be either

                * An :math:`(|E|,)` dense tensor. Each element corresponds to the edge's type ID.
                  Preferred format if ``lowmem == False``.
                * An integer list. The i^th element is the number of edges of the i^th type.
                  This requires the input graph to store edges sorted by their type IDs.
                  Preferred format if ``lowmem == True``.
        norm : torch.Tensor
            Optional edge normalizer tensor. Shape: :math:`(|E|, 1)`.
        edge_feat : torch.Tensor
            Optional argument for the edge features
        efeat_bool : torch.Tensor
            Optional argument to specify which edges in the batch are associated with features for the rest the layer
            behaves as an RGCN

        Returns
        -------
        torch.Tensor
            New node features.

        Notes
        -----
        Under the ``low_mem`` mode, DGL will sort the graph based on the edge types
        and compute message passing one type at a time. DGL recommends sorts the
        graph beforehand (and cache it if possible) and provides the integer list
        format to the ``etypes`` argument. Use DGL's :func:`~dgl.to_homogeneous` API
        to get a sorted homogeneous graph from a heterogeneous graph. Pass ``return_count=True``
        to it to get the ``etypes`` in integer list.
        """
        if isinstance(etypes, th.Tensor):
            if len(etypes) != g.num_edges():
                raise DGLError('"etypes" tensor must have length equal to the number of edges'
                               ' in the graph. But got {} and {}.'.format(
                                   len(etypes), g.num_edges()))
            if self.low_mem and not (feat.dtype == th.int64 and feat.ndim == 1):
                # Low-mem optimization is not enabled for node ID input. When enabled,
                # it first sorts the graph based on the edge types (the sorting will not
                # change the node IDs). It then converts the etypes tensor to an integer
                # list, where each element is the number of edges of the type.
                # Sort the graph based on the etypes
                sorted_etypes, index = th.sort(etypes)
                g = dgl.edge_subgraph(g, index, relabel_nodes=False)
                # Create a new etypes to be an integer list of number of edges.
                pos = th.searchsorted(sorted_etypes, th.arange(self.num_rels, device=g.device))
                num = th.tensor([len(etypes)], device=g.device)
                etypes = (th.cat([pos[1:], num]) - pos).tolist()
                if norm is not None:
                    norm = norm[index]

        with g.local_scope():
            g.srcdata['h'] = feat
            if norm is not None:
                g.edata['norm'] = norm
            if edge_feat is not None:
                g.edata['feat'] = edge_feat
                g.edata['efeat_bool'] = efeat_bool
            g.edata['mask'] = sum(etypes==i for i in self.active_rels)
            if self.self_loop:
                loop_message = matmul_maybe_select(feat[:g.number_of_dst_nodes()],
                                                   self.loop_weight)
            # message passing
            g.update_all(functools.partial(self.message_func, etypes=etypes),
                         fn.sum(msg='msg', out='h'))

            # apply bias and activation
            if 'h' in g.dstdata and len(g.dstdata['h'].shape) == 2:
                node_repr = g.dstdata['h']
            else:
                node_repr = th.zeros((g.number_of_dst_nodes(), self.out_feat),
                                     dtype=feat.dtype,
                                     device=feat.device)
            if self.layer_norm:
                node_repr = self.layer_norm_weight(node_repr)
            if self.bias:
                node_repr = node_repr + self.h_bias
            if self.self_loop:
                node_repr = node_repr + loop_message
            if self.activation:
                node_repr = self.activation(node_repr)
            node_repr = self.dropout(node_repr)
            return node_repr


class NodeTypeEmbedding(nn.Module):
    r"""

   Parameters
   ----------
   in_units : int
       Size of input features
   embed_dim: int
       Size of embedding dimension
    """

    def __init__(self, num_n_types, num_props=2, embed_dim=16, device='cuda'):
        super(NodeTypeEmbedding, self).__init__()

        self.embed_size = embed_dim
        self.num_n_types = num_n_types
        self.num_props = num_props
        self.device = device
        self.ntype_embed = nn.Embedding(self.num_n_types, self.embed_size).to(self.device)

    def forward(self, g, node_ids):
        graph = dgl.block_to_graph(g[0])
        node_tids = graph.ndata[dgl.NTYPE]
        node_types = graph.ntypes
        graph = dgl.add_reverse_edges(dgl.to_homogeneous(graph)).to(self.device)
        h = th.empty((graph.num_nodes(), self.embed_size)).to(self.device)
        for i, n in enumerate(node_types):
            h[graph.ndata[dgl.NTYPE] == i] = self.ntype_embed(node_tids[n].to(self.device))
        with graph.local_scope():
            for i in range(self.num_props):
                graph.ndata['h'] = h
                graph.update_all(fn.copy_src('h', 'm'), fn.sum('m', 'h_new'))
                h = graph.ndata['h_new']

        return h[graph.ndata[dgl.NTYPE] == 0]


import os
import shlex
import subprocess

def num_gpus():
    """Return the number of GPUs available in the current instance.
    Returns:
        int: Number of GPUs available in the current instance.
    """
    try:
        cmd = shlex.split("nvidia-smi --list-gpus")
        output = subprocess.check_output(cmd).decode("utf-8")
        return sum([1 for x in output.split("\n") if x.startswith("GPU ")])
    except (OSError, subprocess.CalledProcessError):
        print("No GPUs detected (normal if no gpus installed)")
        return 0



def get_available_devices():
    """Return the available devices in the current instance
    Returns:
        list[int]: list of available GPU device ids
        [-1] means GPU is not enabled on device and cpu is only available device
    """
    n_gpus = int(os.environ.get('SM_NUM_GPUS', num_gpus()))
    devices = [i for i in range(n_gpus)] if n_gpus > 0 else [-1]
    return devices

def get_device_type(devices):
    """Return the type of available device(s) in the current instance
    Returns:
        str: "cpu" or "gpu"
    """
    return "cpu" if devices[0] == -1 else "cuda"

class objectview(object):
    def __init__(self, d):
        self.__dict__ = d



import torch as th
import torch.nn as nn
from collections.abc import Mapping
from ..mlp import MLP

class MLPFeatureTransformer(nn.Module):
    r"""
   Implementation of MLP feature transformer.

   Transforms a feature dictionary per node type to a feature Tensor per node type by concatenating
   and projecting with an MLP. Optionally, process each feature per node type before concatenating.

   Parameters
   ----------
   in_units : dict[ntype: dict[feat_name: int]]
       Size of input features
   out_units: int
       Size of output units
   num_layers : int or dict[ntype: int]
       number of layers in mlp (Default: 1)
   per_feat_name: bool,
       embed per feature name (Default: False)
   out_unit_per_feat: int
       used if per feature name is set to True (Default: 16)
   dropout_rate: float
       dropout rate, dropout is applied before every layer (Default: 0.)
   hidden_units: int or dict[ntype: int] or dict[ntype: list[int]]
       size of hidden units or list of hidden unit sizes for each hidden layer (Default: None).
       By default set to the in_units
   bias: bool
        whether to use bias in each layer (Default: true)
   """

    def __init__(self,
                 in_units,
                 out_units,
                 num_layers=1,
                 per_feat_name=False,
                 out_unit_per_feat=16,
                 dropout_rate=0,
                 hidden_units=None,
                 bias=True):

        super(MLPFeatureTransformer, self).__init__()

        self.in_units = in_units
        self.out_units = out_units
        self.per_feat_name = per_feat_name
        self.out_units_per_feat = out_unit_per_feat
        self.per_feat_modules = nn.ModuleDict()
        self.ntype_feat_modules = nn.ModuleDict()

        for ntype, feature_dims in in_units.items():
            in_size = 0
            for feature_name, feature_dim in feature_dims.items():
                if self.per_feat_name:
                    self.per_feat_modules[str((ntype, feature_name))] = MLP(feature_dim, out_unit_per_feat,
                                                                            num_layers=1,
                                                                            dropout_rate=dropout_rate, bias=bias)
                    in_size += out_unit_per_feat
                else:
                    in_size += feature_dim

            self.ntype_feat_modules[ntype] = MLP(in_size, out_units, num_layers=num_layers, dropout_rate=dropout_rate,
                                                 bias=bias)


    def forward(self, x):
        h = {}
        for ntype in x:
            ntype_h = []
            for feat_name, feat in x[ntype].items():
                if self.per_feat_name:
                    ntype_h.append(self.per_feat_modules[str((ntype, feat_name))](feat))
                else:
                    ntype_h.append(feat)
            h[ntype] = self.ntype_feat_modules[ntype](th.cat(ntype_h, dim=1))

        return h




"""
Adapted from https://github.com/dmlc/dgl/blob/master/examples/pytorch/han/model_hetero.py
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

import dgl
from dgl.nn.pytorch import GATConv

class SemanticAttention(nn.Module):
    r"""
    Semantic Attention for HAN model.

    Parameters
    ----------
    in_size : int
        Size of input features
    hidden_size : int
        Size of hidden representation
    """
    def __init__(self, in_size, hidden_size):
        super(SemanticAttention, self).__init__()

        self.project = nn.Sequential(
            nn.Linear(in_size, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, 1, bias=False)
        )

    def forward(self, z):
        """

        Parameters
        ----------
        z : th.Tensor
            input. Shape (N, M, D*K)
        Returns
        -------
        th.Tensor
            Aggregate representation for all metapaths. Shape: (N, D * K)
        """
        w = self.project(z).mean(0)                    # (M, 1)
        beta = torch.softmax(w, dim=0)                 # (M, 1)
        beta = beta.expand((z.shape[0],) + beta.shape) # (N, M, 1)

        return (beta * z).sum(1)                       # (N, D * K)

class HANLayer(nn.Module):
    r"""
    HAN layer.

    Parameters
    ----------
    num_metapaths : int
        Number of metapaths
    in_size : int
        Size of input features
    out_size: int
        Size of output representations
    layer_num_heads: int
        number of attention heads (Default: 4)
    dropout: float
        dropout rate for feature and attention weights (Default: 0.)
    activation: Callable
        callable activation function/layer or None, optional. (Default: torch.nn.Functional.elu)
    """
    def __init__(self, num_metapaths, in_size, out_size, layer_num_heads=4, dropout=0., activation=F.elu):
        super(HANLayer, self).__init__()

        # One GAT layer for each meta path based adjacency matrix
        self.gat_layers = nn.ModuleList()
        for i in range(num_metapaths):
            self.gat_layers.append(GATConv(in_size, out_size, layer_num_heads,
                                           dropout, dropout, activation=activation,
                                           allow_zero_in_degree=True))
        self.num_metapaths = num_metapaths

    def forward(self, graph, h):
        """Forward function.
        Compute output of han layer for input node features h
        Parameters
        ----------
        graph : dgl.Graph or dgl.Block
            The graph structure with the same number of nodes as h
        h : th.Tensor or dict[ntype: th.Tensor]
            Nodes embeddings. Shape: (N, D)
        Returns
        -------
        th.Tensor
            Representation for edge between u and v. Shape: (B, out)
        """
        metapath_embeddings = []
        for i, g in enumerate(graph):
            metapath_embeddings.append(self.gat_layers[i](g, h[i]).flatten(1))
        return  metapath_embeddings

class HANEncoder(nn.Module):
    r"""
    HAN Encoder Module.

    Paper link: https://github.com/Jhy1993/Representation-Learning-on-Heterogeneous-Graph

    Current implementation supports only node classification tasks

    Parameters
    ----------
    num_metapaths : list[str] or list[tuple[str]]
        Metapath in the form of a list of edge types
    in_size : int
        Size of input features
    hidden_size: int
        Size of output representations
    num_layers: int
        number of HAN layers
    num_heads: int or list[int]
        number of attention heads for each layer (Default: 4)
    dropout: float
        dropout rate for feature and attention weights (Default: 0.)
    activation: Callable
        callable activation function/layer or None, optional. (Default: torch.nn.Functional.elu)
    """
    def __init__(self, num_metapaths, in_size, hidden_size, num_layers, num_heads=4, dropout=0., activation=F.elu):
        super(HANEncoder, self).__init__()

        self.layers = nn.ModuleList()
        if isinstance(num_heads, int):
            num_heads = [num_heads] * (num_layers)
        else:
            assert len(num_heads) == num_layers, "Mismatch between attention heads for each layer and number of layers"
        self.layers.append(HANLayer(num_metapaths, in_size, hidden_size, num_heads[0], dropout, activation))
        for l in range(1, num_layers):
            self.layers.append(HANLayer(num_metapaths, hidden_size * num_heads[l-1],
                                        hidden_size, num_heads[l], dropout, activation))
        self.semantic_attention = SemanticAttention(in_size=hidden_size * num_heads[l], hidden_size=hidden_size)
        self.project = nn.Linear(hidden_size * num_heads[-1], hidden_size)

    def forward(self, g, h):
        r"""
        Compute the node representations on initial node features with graph g

        Parameters
        ----------
        g : list[dgl.Graph/dgl.Block] or list[list[dgl.Graph/dgl.Block]]
            The list of graphs for each metapath. For more than one layers
            each entry in the list should be list of message flow graphs for each metapath
        h: list[th.Tensor]
            The list of initial node features for the input nodes for each metapath

        """
        for l, gnn in enumerate(self.layers):
            mfg = [block[l] for block in g] if type(g[0]) == list else g
            h = gnn(mfg, h)
        semantic_embeddings = self.semantic_attention(torch.stack(h, dim=1))  # (N, D * K)
        return self.project(semantic_embeddings)


import boto3
from datetime import datetime

try:
    boto_session = boto3.DEFAULT_SESSION or boto3.Session()
    region_name = boto_session.region_name
    cloudwatch = boto3.client(
        service_name='logs',
        region_name=region_name)
except:
    print("Could not initialize boto client")

def get_log_stream_name(log_group_name, log_stream_prefix):
    kwargs = {'logGroupName': log_group_name,
              'logStreamNamePrefix': log_stream_prefix,
              'orderBy': 'LogStreamName',
              'descending': True}
    log_streams = cloudwatch.describe_log_streams(**kwargs)
    # assumes that only one log stream should be returned since neptune ml job names are unique and are not prefixes
    # of other job names
    if 'logStreams' in log_streams:
        if log_streams['logStreams']:
            return log_streams['logStreams'][0]['logStreamName']
    else:
        print('Log Stream not found for log group %s, and log stream prefix %s' % (log_group_name, log_stream_prefix))

    return None

def print_logs(log_group_name, log_stream_name, start_time):
    kwargs = {'logGroupName': log_group_name,
              'logStreamName': log_stream_name,
              'startTime': start_time,
              'startFromHead': True}

    lastTimestamp = start_time
    while True:
        logEvents = cloudwatch.get_log_events(**kwargs)
        nextToken = logEvents['nextForwardToken']
        events = logEvents['events']

        for event in events:
            lastTimestamp = event['timestamp']
            timestamp = datetime.utcfromtimestamp(lastTimestamp / 1000.0).isoformat()
            print('[%s] %s' % ((timestamp + ".000")[:23] + 'Z', event['message']))

        if nextToken and kwargs.get('nextToken') != nextToken:
            kwargs['nextToken'] = nextToken
        else:
            break
    return lastTimestamp


import boto3
import os
from copy import copy
from botocore.session import get_session
from graph_notebook.neptune.client import Client, ClientBuilder, NEPTUNE_CONFIG_HOST_IDENTIFIERS, \
    is_allowed_neptune_host
from graph_notebook.configuration.generate_config import (generate_default_config, DEFAULT_CONFIG_LOCATION,
                                                          AuthModeEnum, Configuration)
from graph_notebook.configuration.get_config import get_config
from urllib.parse import urlparse

from .utils import s3_utils, sagemaker_utils, batch_utils, neptune_utils

import numpy as np
import pickle
import time

try:
    DEFAULT_REGION = boto3.Session().region_name
except:
    print("Could not get region from boto session. Use us-east-1 by default")
    DEFAULT_REGION = 'us-east-1'
DEFAULT_PROCESSING_CONFIG_FILE_NAME = 'training-data-configuration.json'
DEFAULT_TRAINING_CONFIG_FILE_NAME = 'model-hpo-configuration.json'
DEFAULT_EMBEDDING_FILE_NAME = 'entity.npy'
DEFAULT_PREDICTION_FILE_NAME = 'result.npz'
DEFAULT_TRAINING_MAPPING_INFO = 'mapping.info'
DEFAULT_PROCESSING_MAPPING_INFO = 'info.pkl'
DEFAULT_TRAINING_MAPPING_KEY = 'node2id'
DEFAULT_PROCESSING_MAPPING_KEY = 'node_id_map'
DEFAULT_ENTITY_MAPPING_KEY = 'node2gid'
DEFAULT_MAX_HPO_NUM = 2
DEFAULT_MAX_PARALLEL_HPO = 2
DEFAULT_EVAL_METRIC_FILE = "eval_metrics_info.json"
DEFAULT_TEST_METRIC_FILE = "test_metrics_info.json"
DEFAULT_TRAIN_INSTANCE_REC = "train_instance_recommendation.json"
DEFAULT_INFER_INSTANCE_REC = "infer_instance_recommendation.json"


class NeptuneMLClient():

    def __init__(self, host=None, port=8182, auth_mode='DEFAULT', use_ssl=False, region=DEFAULT_REGION,
                 load_from_s3_arn='', proxy_host='', proxy_port=8182, neptune_hosts=NEPTUNE_CONFIG_HOST_IDENTIFIERS,
                 config_location=DEFAULT_CONFIG_LOCATION, neptune_iam_role_arn='', sagemaker_iam_role_arn='',
                 subnets=None, security_group_ids=None, volume_encryption_kms_key=None,
                 s3_output_encryption_kms_key=None,
                 export_service_uri=None):
        """
        Creates a NeptuneMLClient for interacting NeptuneML model management APIs

        :param host: (str) the host url to form a connection with.
        :param port: (int) the port to use when creating a connection. Default 8182
        :param auth_mode: (str) type of authentication the cluster being connected to is using. Can be DEFAULT or IAM
        :param use_ssl: (bool) whether to make connections to the created endpoint with ssl or not. Default True
        :param region: (str) aws region your Neptune cluster is in. Default
        :param load_from_s3_arn: (str) arn of role to use for bulk loader.
        :param proxy_host: (str) the proxy host url to route a connection through.
        :param proxy_port: (int) the proxy port to use when creating proxy connection. Default 8182
        :param neptune_hosts: list of host snippets to use for identifying neptune endpoints. Default ["amazonaws.com"]
        :param config_location: (str) location of existing Neptune db client config. Default '~/graph_notebook_config.json'
        :param neptune_iam_role_arn: (str) arn of role used by neptune to for neptune ml commands
        :param sagemaker_iam_role_arn: (str) arn of role to use within sagemaker jobs
        :param subnets: list of the IDs of the subnets in the Neptune VPC
        :param security_group_ids: list of The VPC security group IDs
        :param volume_encryption_kms_key: (str) AWS KMS Key to be used for encypting the storage volume on all SageMaker jobs by the client
        :param s3_output_encryption_kms_key: (str) AWS KMS Key to be used for encypting s3 objects by all SageMaker jobs by the client
        :param export_service_uri: (str) url of Neptune export service
        """
        self.client = None
        self.neptune_cfg_allowlist = copy(neptune_hosts)

        if host is not None:
            auth_mode = auth_mode if auth_mode in [auth.value for auth in AuthModeEnum] else AuthModeEnum.DEFAULT.value
            config = generate_config(host, port, AuthModeEnum(auth_mode), use_ssl, load_from_s3_arn, region,
                                     proxy_host, proxy_port, neptune_hosts=neptune_hosts)
        else:
            self.config_location = config_location or os.getenv('GRAPH_NOTEBOOK_CONFIG', DEFAULT_CONFIG_LOCATION)
            config = get_config(self.config_location, neptune_hosts=self.neptune_cfg_allowlist)

        self._config = config
        self._generate_client_from_config(config)

        if not neptune_iam_role_arn:
            neptune_iam_role_arn = neptune_utils.get_neptune_ml_iam_role_parameter_group(self.get_host())

        self._register_security_params(sagemaker_iam_role_arn, neptune_iam_role_arn, subnets, security_group_ids,
                                       volume_encryption_kms_key, s3_output_encryption_kms_key)

        self._export_service_host = None
        if export_service_uri is not None:
            path = urlparse(export_service_uri.rstrip())
            self._export_service_host = path.hostname + "/v1"
        else:
            self._export_service_host = self.get_export_service_host()
            assert self._export_service_host is not None, "Export service is required for Neptune ML"

    def _generate_client_from_config(self, config):
        if self.client:
            self.client.close()

        is_neptune_host = is_allowed_neptune_host(hostname=config.host, host_allowlist=self.neptune_cfg_allowlist)
        assert is_neptune_host, "Neptune ML only works with Amazon NeptuneDB"

        self.builder = ClientBuilder() \
            .with_host(config.host) \
            .with_port(config.port) \
            .with_region(config.aws_region) \
            .with_tls(config.ssl) \
            .with_proxy_host(config.proxy_host) \
            .with_proxy_port(config.proxy_port) \
            .with_sparql_path(config.sparql.path)
        if config.auth_mode == AuthModeEnum.IAM:
            self.builder = self.builder.with_iam(get_session())
        if self.neptune_cfg_allowlist != NEPTUNE_CONFIG_HOST_IDENTIFIERS:
            self.builder = self.builder.with_custom_neptune_hosts(self.neptune_cfg_allowlist)

        self.client = self.builder.build()

    def _register_security_params(self, sagemaker_iam_role_arn, neptune_iam_role_arn, subnets, security_group_ids,
                                  volume_encryption_kms_key, s3_output_encryption_kms_key):
        self.sagemaker_iam_role_arn = sagemaker_iam_role_arn
        self.neptune_iam_role_arn = neptune_iam_role_arn
        self.subnets = subnets
        self.security_group_ids = security_group_ids
        self.volume_encryption_kms_key = volume_encryption_kms_key
        self.s3_output_encryption_kms_key = s3_output_encryption_kms_key

    def check_enabled(self):
        """
        check whether the current environment and neptune cluster is configured for Neptune ML
        """
        try:
            assert self.get_host() is not None
            data_processing_job = self.list_data_processing_jobs()
            assert self.get_export_service_host() is not None
            neptune_utils.get_neptune_ml_iam_role_parameter_group(self.get_host())
            print("This Neptune cluster is configured to use Neptune ML")
        except Exception as ex:
            print(ex)
            print('''This Neptune cluster \033[1mis not\033[0m configured to use Neptune ML.
                Please configure the cluster according to the Amazon Neptune ML documentation before proceeding.''')

    def get_export_service_host(self):
        """
        get the host url for the Neptune ML export service. Assumes it's set by user or in .bashrc
        """

        if self._export_service_host is not None:
            return self._export_service_host
        try:
            with open('/home/ec2-user/.bashrc') as f:
                data = f.readlines()
            for d in data:
                if str.startswith(d, 'export NEPTUNE_EXPORT_API_URI'):
                    parts = d.split('=')
                    if len(parts) == 2:
                        path = urlparse(parts[1].rstrip())
                        return path.hostname + "/v1"
        except:
            print("Unable to determine the Neptune Export Service Endpoint. You will need to this assign this manually")
        return None

    def get_host(self):
        if self._config is not None:
            return self._config.host
        else:
            return None

    def get_port(self):
        if self._config is not None:
            return self._config.port
        else:
            return None

    def get_iam(self):
        if self._config is not None:
            return self._config.auth_mode == AuthModeEnum.IAM
        else:
            return None

    def _add_security_params(self, params, sagemaker_iam_role_arn, neptune_iam_role_arn, subnets, security_group_ids,
                             volume_encryption_kms_key, s3_output_encryption_kms_key):

        sagemaker_iam_role_arn = sagemaker_iam_role_arn or self.sagemaker_iam_role_arn
        if sagemaker_iam_role_arn:
            params['sagemakerIamRoleArn'] = sagemaker_iam_role_arn
        neptune_iam_role_arn = neptune_iam_role_arn or self.neptune_iam_role_arn
        if neptune_iam_role_arn:
            params['neptuneIamRoleArn'] = neptune_iam_role_arn
        subnets = subnets or self.subnets
        if subnets:
            if all([isinstance(subnet, str) for subnet in subnets]):
                params['subnets'] = subnets
            else:
                print('Ignoring subnets, list does not contain all strings.')
        security_group_ids = security_group_ids or self.security_group_ids
        if security_group_ids:
            if all([isinstance(security_group_id, str) for security_group_id in security_group_ids]):
                params['securityGroupIds'] = security_group_ids
            else:
                print('Ignoring security group IDs, list does not contain all strings.')
        volume_encryption_kms_key = volume_encryption_kms_key or self.volume_encryption_kms_key
        if volume_encryption_kms_key:
            params['volumeEncryptionKMSKey'] = volume_encryption_kms_key
        s3_output_encryption_kms_key = s3_output_encryption_kms_key or self.s3_output_encryption_kms_key
        if s3_output_encryption_kms_key:
            params['s3OutputEncryptionKMSKey'] = s3_output_encryption_kms_key

        return params

    def create_data_export_job(self,
                               command=None,
                               outputS3Path=None,
                               jobSize=None,
                               exportProcessParams=None,
                               additionalParams=None,
                               export_url=None,
                               export_iam=True,
                               ssl=True,
                               wait=False,
                               params={}):
        """
        Create a new Neptune ML data export job. See https://docs.aws.amazon.com/neptune/latest/userguide/export-parameters.html for details on the parameters
        If wait is set to True, this waits for the jobs to complete and streams logs to std out.

        :param command: determines whether to export property-graph data or RDF data.
        :param outputS3Path: The URI of an Amazon S3 location to which the exported files can be published (Required)
        :param jobSize: the size of the export job you are starting. options - ["small", "medium", "large", "xlarge"]
        :param exportProcessParams: a JSON dict that contains parameters that you use to control the export process itself see: https://docs.aws.amazon.com/neptune/latest/userguide/export-params-fields.html
        :param additionalParams: additionalParams top-level parameter is a JSON object that contains parameters you can use to control actions that are applied to the data after it has been exported see: https://docs.aws.amazon.com/neptune/latest/userguide/machine-learning-data-export.html#machine-learning-additionalParams
        :param export_url: url of the Neptune export service. See get_export_service_host()
        :param export_iam: Whether to use iam authentication for export service calls. default True
        :param ssl: Whether to enable SSL. default True
        :param wait: Whether to run job synchronously and wait for job to complete. default False
        :param params: a JSON dict for export parameters to use instead of the kwargs above to simplify parameter management. See https://docs.aws.amazon.com/neptune/latest/userguide/export-parameters.html

        :return: a JSON dict with the job response
        """
        if params:
            assert "command" in params, "'command' must be in params and be one of export_pg or export_rdf"
            assert "outputS3Path" in params, "'outputS3Path' must be in params"
        else:
            assert outputS3Path is not None, "outputS3path is required"
            params = {}
            params['outputS3Path'] = outputS3Path
            if command is not None:
                params['command'] = command
            if jobSize is not None:
                params['jobSize'] = jobSize
            if exportProcessParams is not None:
                params['params'] = exportProcessParams
            if additionalParams is not None:
                params['additionalParams'] = additionalParams

        if export_iam:
            self.builder = self.builder.with_iam(get_session())
        export_client = self.builder.build()

        export_url = export_url or self.get_export_service_host()
        export_job_response = export_client.export(export_url, params, ssl)
        export_job_response.raise_for_status()
        export_job = export_job_response.json()
        if wait == True:
            batch_utils.wait_for_job(export_job["jobId"])
            return self.describe_data_export_job(export_job["jobId"])
        return export_job

    def describe_data_export_job(self, id, export_url=None, export_iam=True, ssl=True):
        """
        Describe details of a Neptune ML data export job

        :param id: id of the export job
        :param export_url: url of the Neptune export service. See get_export_service_host()
        :param export_iam: Whether to use iam authentication for export service calls. default True
        :param ssl: Whether to enable SSL. default True

        :return: a JSON dict with the job response
        """
        if export_iam:
            self.builder = self.builder.with_iam(get_session())
        export_client = self.builder.build()

        export_url = export_url or self.get_export_service_host()
        export_status = export_client.export_status(export_url, id, ssl)
        export_status.raise_for_status()
        export_job = export_status.json()
        return export_job

    def get_training_data_configuration(self, export_job_id):
        """
        Get training data configuration from the file that is created by a Neptune ML data export job

        :param export_job_id: id of the export job

        :return: a JSON dict with the configuration
        """
        export_job = self.describe_data_export_job(export_job_id)
        return s3_utils.get_config(export_job['outputS3Uri'], DEFAULT_PROCESSING_CONFIG_FILE_NAME)

    def create_data_processing_job(self,
                                   id=None,
                                   previousDataProcessingJobId=None,
                                   inputDataS3Location=None,
                                   processedDataS3Location=None,
                                   sagemakerIamRoleArn=None,
                                   neptuneIamRoleArn=None,
                                   processingInstanceType=None,
                                   processingInstanceVolumeSizeInGB=None,
                                   processingTimeOutInSeconds=None,
                                   modelType=None,
                                   configFileName=None,
                                   subnets=None,
                                   securityGroupIds=None,
                                   volumeEncryptionKMSKey=None,
                                   s3OutputEncryptionKMSKey=None,
                                   trainingDataConfiguration=None,
                                   params={},
                                   wait=False):
        """
        Create a new Neptune ML data processing job. See https://docs.aws.amazon.com/neptune/latest/userguide/machine-learning-api-dataprocessing.html#machine-learning-api-dataprocessing-create-job for details
        If wait is set to True, this waits for the jobs to complete and streams logs to std out.

        :param id: A unique identifier for the new job. if None a uuid based id is autogenerated
        :param previousDataProcessingJobId: The job ID of a completed data processing job run on an earlier version of the data for incremental offline inference workflows
        :param inputDataS3Location: The URI of the Amazon S3 location where you want SageMaker to download the data needed to run the data processing job (Required)
        :param processedDataS3Location: The URI of the Amazon S3 location where you want SageMaker to save the results of a data processing job (Required)
        :param sagemakerIamRoleArn: The Amazon Resource Name (ARN) of an IAM role that is passed to SageMaker and can be assumed by SageMaker to perform tasks on your behalf.
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker. Note: This must be listed in your DB cluster parameter group or an error will occur.
        :param processingInstanceType: The type of ML instance used during data processing. Its memory should be large enough to hold the processed dataset.
        :param processingInstanceVolumeSizeInGB: The disk volume size of the processing instance. Both input data and processed data are stored on disk, so the volume size must be large enough to hold both data sets.
        :param processingTimeOutInSeconds: Timeout in seconds for the data processing job.
        :param modelType: One of the two model types that Neptune ML currently supports: heterogeneous graph models (heterogeneous), and knowledge graph (kge)
        :param configFileName: A data specification file that describes how to load the exported graph data for training. The file is automatically generated by the Neptune export toolkit default training-data-configuration.json
        :param subnets: The IDs of the subnets in the Neptune VPC
        :param securityGroupIds: The VPC security group IDs.
        :param volumeEncryptionKMSKey: The AWS Key Management Service (AWS KMS) key that SageMaker uses to encrypt data on the storage volume attached to the ML compute instances that run the processing job
        :param s3OutputEncryptionKMSKey: The AWS Key Management Service (AWS KMS) key that SageMaker uses to encrypt the output of the training job.
        :param trainingDataConfiguration: a JSON dict with the configuration to load the exported graph data for training. If not None, the configuration is uploaded to S3 and overwrites the file at inputDataS3Location/configFileName
        :param params: a JSON dict for data processing parameters to use instead of the kwargs above to simplify parameter management
        :param wait: Whether to run job synchronously and wait for job to complete. default False

        :return: a JSON dict with the job response
        """

        if trainingDataConfiguration is not None:
            if type(trainingDataConfiguration) == str:
                assert os.path.exists(trainingDataConfiguration), "trainingDataConfiguration file not found at path {}".format(trainingDataConfiguration)
                s3_utils.upload_config(trainingDataConfiguration, inputDataS3Location)
            else:
                assert type(trainingDataConfiguration) == dict
                configFileName = configFileName or params.get('configFileName', DEFAULT_PROCESSING_CONFIG_FILE_NAME)
                s3_utils.upload_config(trainingDataConfiguration, inputDataS3Location, s3_file_name=configFileName)

        if params:
            assert 'inputDataS3Location' in params or inputDataS3Location, "inputDataS3Location must be provided"
            assert 'processedDatas3Location' in params or processedDataS3Location, "processedDataS3Location must be provided"
            inputDataS3Location = inputDataS3Location or params['inputDataS3Location']
            processedDataS3Location = processedDataS3Location or params['processedDataS3Location']
        else:
            assert inputDataS3Location is not None, "inputDataS3Location must be provided"
            assert processedDataS3Location is not None, "processedDataS3Location must be provided"
            params = {}
            if id:
                params['id'] = id
            if previousDataProcessingJobId:
                params['previousDataProcessingJobId'] = previousDataProcessingJobId
            if modelType:
                params['modelType'] = modelType
            if configFileName:
                params['configFileName'] = configFileName
            self._add_security_params(params, sagemakerIamRoleArn, neptuneIamRoleArn, subnets, securityGroupIds,
                                      volumeEncryptionKMSKey, s3OutputEncryptionKMSKey)

        data_processing_job_response = self.client.dataprocessing_start(inputDataS3Location, processedDataS3Location,
                                                                        **params)
        data_processing_job_response.raise_for_status()
        data_processing_job = data_processing_job_response.json()
        if wait:
            neptuneIamRoleArn = params.get("neptuneIamRoleArn", neptuneIamRoleArn)
            data_processing_status = self.describe_data_processing_job(data_processing_job["id"],
                                                                       neptuneIamRoleArn=neptuneIamRoleArn)
            sagemaker_utils.wait_for_job(data_processing_status["processingJob"]["name"])

        return data_processing_job

    def describe_data_processing_job(self, id, neptuneIamRoleArn=None, verbose=False):
        """

        :param id: identifier of the data processing job
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker
        :param verbose: Whether to include granular job details from SageMaker Processing job

        :return: a JSON dict with the job response
        """
        neptuneIamRoleArn = neptuneIamRoleArn or self.neptune_iam_role_arn
        if neptuneIamRoleArn is not None:
            data_processing_status = self.client.dataprocessing_job_status(id, neptune_iam_role_arn=neptuneIamRoleArn)
        else:
            data_processing_status = self.client.dataprocessing_job_status(id)

        data_processing_status.raise_for_status()
        data_processing_job = data_processing_status.json()

        if verbose:
            data_processing_job["processingJobDetails"] = sagemaker_utils.get_processing_job_details(
                data_processing_job["processingJob"]["name"])

        return data_processing_job

    def list_data_processing_jobs(self, maxItems=10, neptuneIamRoleArn=None):
        """
        List data processing jobs associated with the current DB cluster config

        :param maxItems: number of Items to return at once
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker

        :return: a JSON list with the job response
        """
        neptuneIamRoleArn = neptuneIamRoleArn or self.neptune_iam_role_arn
        if neptuneIamRoleArn is not None:
            list_result = self.client.dataprocessing_list(max_items=maxItems, neptune_iam_role_arn=neptuneIamRoleArn)
        else:
            list_result = self.client.dataprocessing_list(max_items=maxItems)
        list_result.raise_for_status()
        return list_result.json()

    def stop_data_processing_job(self, id, clean=False, neptuneIamRoleArn=None):
        """
        Stop an existing data processing job.

        :param id: identifier of the data processing job
        :param clean: whether to delete any S3 artifacts created by the stopped job
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker
        """
        neptuneIamRoleArn = neptuneIamRoleArn or self.neptune_iam_role_arn
        if neptuneIamRoleArn is not None:
            stop_result = self.client.modeltransform_stop(id, clean=clean, neptune_iam_role_arn=neptuneIamRoleArn)
        else:
            stop_result = self.client.modeltransform_stop(id, clean=clean)
        stop_result.raise_for_status()

    def get_model_hpo_configuration(self, data_processing_job_id):
        """
        Get model and hpo configuration from the file that is created by a Neptune ML data processing job

        :param data_processing_job_id: id of the data processing job

        :return: a JSON dict with the configuration
        """
        data_processing_job = self.describe_data_processing_job(data_processing_job_id)
        return s3_utils.get_config(data_processing_job["processingJob"]["outputLocation"],
                                   DEFAULT_TRAINING_CONFIG_FILE_NAME)

    def get_model_training_instance_recommendation(self, data_processing_job_id):
        """
        Get configuration for recommended model training instance from the file that is created by a Neptune ML data processing job

        :param data_processing_job_id: id of the data processing job
        :return:
        """
        data_processing_job = self.describe_data_processing_job(data_processing_job_id)
        return s3_utils.get_config(data_processing_job["processingJob"]["outputLocation"], DEFAULT_TRAIN_INSTANCE_REC)

    def create_model_training_job(self,
                                  id=None,
                                  dataProcessingJobId=None,
                                  trainModelS3Location=None,
                                  previousModelTrainingJobId=None,
                                  sagemakerIamRoleArn=None,
                                  neptuneIamRoleArn=None,
                                  modelName=None,
                                  baseProcessingInstanceType=None,
                                  trainingInstanceType=None,
                                  trainingInstanceVolumeSizeInGB=None,
                                  trainingTimeOutInSeconds=None,
                                  maxHPONumberOfTrainingJobs=None,
                                  maxHPOParallelTrainingJobs=None,
                                  subnets=None,
                                  securityGroupIds=None,
                                  volumeEncryptionKMSKey=None,
                                  s3OutputEncryptionKMSKey=None,
                                  enableManagedSpotTraining=None,
                                  customModelTrainingParameters=None,
                                  modelHPOConfiguration=None,
                                  params={},
                                  wait=False):
        """
        Create a new Neptune ML model training job. See https://docs.aws.amazon.com/neptune/latest/userguide/machine-learning-api-modeltraining.html#machine-learning-api-modeltraining-create-job for details
        If wait is set to True, this waits for the jobs to complete and streams logs to std out.

        :param id: A unique identifier for the new job. if None a uuid based id is autogenerated
        :param dataProcessingJobId: The job Id of the completed data-processing job that has created the data for this training job. (Required)
        :param trainModelS3Location: The location in Amazon S3 where the model artifacts are to be stored (Required)
        :param previousModelTrainingJobId: The job ID of a completed model-training job that you want to update incrementally with this training job
        :param sagemakerIamRoleArn: The Amazon Resource Name (ARN) of an IAM role that is passed to SageMaker and can be assumed by SageMaker to perform tasks on your behalf.
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker. Note: This must be listed in your DB cluster parameter group or an error will occur.
        :param modelName: The model type for training. By default the ML model is automatically based on the modelType used in data processing, but you can specify a different model type here. For heterogeneous graphs: rgcn. For kge graphs: transe, distmult, or rotate. For a custom model implementation: custom.
        :param baseProcessingInstanceType: The type of ML instance used in preparing and managing training of ML models.
        :param trainingInstanceType: The type of ML instance used for model training. All Neptune ML models support CPU and GPU training
        :param trainingInstanceVolumeSizeInGB: The disk volume size of the training instance. Both input data and the output model are stored on disk, so the volume size must be large enough to hold both data sets.
        :param trainingTimeOutInSeconds: Timeout in seconds for the training job.
        :param maxHPONumberOfTrainingJobs: Maximum total number of training jobs to start for the hyperparameter tuning job.
        :param maxHPOParallelTrainingJobs: Maximum number of parallel training jobs to start for the hyperparameter tuning job.
        :param subnets: The IDs of the subnets in the Neptune VPC
        :param securityGroupIds: The VPC security group IDs.
        :param volumeEncryptionKMSKey: The AWS Key Management Service (AWS KMS) key that SageMaker uses to encrypt data on the storage volume attached to the ML compute instances that run the processing job
        :param s3OutputEncryptionKMSKey: The AWS Key Management Service (AWS KMS) key that SageMaker uses to encrypt the output of the training job.
        :param enableManagedSpotTraining: Optimizes the cost of training machine learning models by using Amazon Elastic Compute Cloud spot instances
        :param customModelTrainingParameters: The configuration for custom model training. This is a JSON object with the following fields sourceDirectory, sourceS3DirectoryPath, trainingEntryPointScript, transformEntryPointScript
        :param modelHPOConfiguration: a JSON dict with the configuration for managing model training and HPO. If not None, the configuration is uploaded to S3 and overwrites the file at inputDataS3Location/model-hpo-configuration.json
        :param params: a JSON dict for model training parameters to use instead of the kwargs above to simplify parameter management
        :param wait: Whether to run job synchronously and wait for job to complete. default False

        :return: a JSON dict with the job response
        """
        assert dataProcessingJobId is not None or "dataProcessingJobId" in params, "dataProcessingId is required"
        assert trainModelS3Location is not None or "trainModelS3ocation" in params, "trainModelS3Location is required"
        if modelHPOConfiguration is not None:
            s3_train_input_uri = self.describe_data_processing_job(data_processing_job["id"],
                                                                   neptuneIamRoleArn=neptuneIamRoleArn)[
                "processingJob"]["outputLocation"]
            if type(modelHPOConfiguration) == str:
                assert os.path.exists(
                    modelHPOConfiguration), "trainingDataConfiguration file not found at path {}".format(
                    trainingDataConfiguration)
                s3_utils.upload_config(modelHPOConfiguration, s3_train_input_uri)
            else:
                assert type(trainingDataConfiguration) == dict
                s3_utils.upload_config(modelHPOConfiguration, s3_train_input_uri,
                                       s3_file_name=DEFAULT_TRAINING_CONFIG_FILE_NAME)

        if customModelTrainingParameters is not None:
            customModelTrainingParameters = dict(customModelTrainingParameters)
            assert 'sourceS3DirectoryPath' in customModelTrainingParameters, "sourceS3DirectoryPath must be provided in customModelTrainingParameters"
            if 'sourceDirectory' in customModelTrainingParameters:
                source_directory = customModelTrainingParameters.pop('sourceDirectory')
                s3_utils.upload(source_directory, customModelTrainingParameters['sourceS3DirectoryPath'])

        if params:
            dataProcessingJobId = dataProcessingJobId or params['dataProcessingJobId']
            trainModelS3Location = trainModelS3Location or params['trainModelS3Location']
            maxHPONumberOfTrainingJobs = maxHPONumberOfTrainingJobs or params['maxHPONumberOfTrainingJobs']
            maxHPOParallelTrainingJobs = maxHPOParallelTrainingJobs or params['maxHPOParallelTrainingJobs']
        else:
            maxHPONumberOfTrainingJobs = maxHPONumberOfTrainingJobs or DEFAULT_MAX_HPO_NUM
            maxHPOParallelTrainingJobs = maxHPOParallelTrainingJobs or DEFAULT_MAX_PARALLEL_HPO
            params = {}
            if id:
                params['id'] = id
            if previousModelTrainingJobId:
                params['previousModelTrainingJobId'] = previousModelTrainingJobId
            if modelName:
                params['modelName'] = modelName
            if baseProcessingInstanceType:
                params['baseProcessingInstanceType'] = baseProcessingInstanceType
            if trainingInstanceType:
                params['trainingInstanceType'] = trainingInstanceType
            if trainingInstanceVolumeSizeInGB:
                params['trainingInstanceVolumeSizeInGB'] = trainingInstanceVolumeSizeInGB
            if trainingTimeOutInSeconds:
                params['trainingTimeOutInSeconds'] = trainingTimeOutInSeconds
            if enableManagedSpotTraining:
                params['enableManagedSpotTraining'] = enableManagedSpotTraining
            if customModelTrainingParameters:
                params['customModelTrainingParameters'] = customModelTrainingParameters

            self._add_security_params(params, sagemakerIamRoleArn, neptuneIamRoleArn, subnets, securityGroupIds,
                                      volumeEncryptionKMSKey, s3OutputEncryptionKMSKey)

        model_training_job_response = self.client.modeltraining_start(dataProcessingJobId, trainModelS3Location,
                                                                      maxHPONumberOfTrainingJobs,
                                                                      maxHPOParallelTrainingJobs, **params)
        model_training_job_response.raise_for_status()
        model_training_job = model_training_job_response.json()
        if wait:
            neptuneIamRoleArn = params.get("neptuneIamRoleArn", neptuneIamRoleArn)
            model_training_status = self.describe_model_training_job(model_training_job["id"],
                                                                     neptuneIamRoleArn=neptuneIamRoleArn)
            sagemaker_utils.wait_for_job(model_training_status["processingJob"]["name"])

        return model_training_job

    def describe_model_training_job(self, id, neptuneIamRoleArn=None, verbose=False):
        """

        :param id: identifier of the model training job
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker
        :param verbose: Whether to include granular job details from SageMaker Processing job

        :return: a JSON dict with the job response
        """
        neptuneIamRoleArn = neptuneIamRoleArn or self.neptune_iam_role_arn
        if neptuneIamRoleArn is not None:
            model_training_status = self.client.modeltraining_job_status(id, neptune_iam_role_arn=neptuneIamRoleArn)
        else:
            model_training_status = self.client.modeltraining_job_status(id)

        model_training_status.raise_for_status()
        model_training_job = model_training_status.json()

        if verbose:
            model_training_job["processingJobDetails"] = sagemaker_utils.get_processing_job_details(
                model_training_job["processingJob"]["name"])
            model_training_job[
                "hyperparameterTuningJobDetails"] = sagemaker_utils.get_hyperparameter_tuning_job_details(
                model_training_job["hpoJob"]["name"])

        return model_training_job

    def list_model_training_jobs(self, maxItems=10, neptuneIamRoleArn=None):
        """
        List model training jobs associated with the current DB cluster config

        :param maxItems: number of Items to return at once
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker

        :return: a JSON list with the job response
        """
        neptuneIamRoleArn = neptuneIamRoleArn or self.neptune_iam_role_arn
        if neptuneIamRoleArn is not None:
            list_result = self.client.modeltraining_list(max_items=maxItems, neptune_iam_role_arn=neptuneIamRoleArn)
        else:
            list_result = self.client.modeltraining_list(max_items=maxItems)
        list_result.raise_for_status()
        return list_result.json()

    def stop_model_training_job(self, id, clean=False, neptuneIamRoleArn=None):
        """
        Stop an existing model training job.

        :param id: identifier of the model training job
        :param clean: whether to delete any S3 artifacts created by the stopped job
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker
        """
        neptuneIamRoleArn = neptuneIamRoleArn or self.neptune_iam_role_arn
        if neptuneIamRoleArn is not None:
            stop_result = self.client.modeltraining_stop(id, clean=clean, neptune_iam_role_arn=neptuneIamRoleArn)
        else:
            stop_result = self.client.modeltraining_stop(id, clean=clean)
        stop_result.raise_for_status()

    def create_model_transform_job(self,
                                   id=None,
                                   dataProcessingJobId=None,
                                   mlModelTrainingJobId=None,
                                   trainingJobName=None,
                                   modelTransformOutputS3Location=None,
                                   sagemakerIamRoleArn=None,
                                   neptuneIamRoleArn=None,
                                   baseProcessingInstanceType=None,
                                   baseProcessingInstanceVolumeSizeInGB=None,
                                   subnets=None,
                                   securityGroupIds=None,
                                   volumeEncryptionKMSKey=None,
                                   s3OutputEncryptionKMSKey=None,
                                   customModelTransformParameters=None,
                                   params={},
                                   wait=False):
        """
        Create a new Neptune ML model transform job. See https://docs.aws.amazon.com/neptune/latest/userguide/machine-learning-api-modeltransform.html#machine-learning-api-modeltransform-create-job for details
        If wait is set to True, this waits for the jobs to complete and streams logs to std out.

        :param id: A unique identifier for the new job. if None a uuid based id is autogenerated
        :param dataProcessingJobId: The job Id of a completed data-processing job as input to the model transform job
        :param mlModelTrainingJobId: The job Id of a completed model-training job with a generated best model that is used for the transform
        :param trainingJobName: The name of a completed sagemaker training job. Only applicable when you want a different model other than the best model selected by Neptune ML for the transform. ignored if mlModelTrainingJobId is passed in
        :param modelTransformOutputS3Location: The location in Amazon S3 where new model artifacts and predictions are to be stored (Required)
        :param sagemakerIamRoleArn: The Amazon Resource Name (ARN) of an IAM role that is passed to SageMaker and can be assumed by SageMaker to perform tasks on your behalf.
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker. Note: This must be listed in your DB cluster parameter group or an error will occur.
        :param baseProcessingInstanceType: The type of ML instance used in preparing and managing training of ML models.
        :param baseProcessingInstanceVolumeSizeInGB: The disk volume size of the transform instance. Both input data and the output model are stored on disk, so the volume size must be large enough to hold both data sets.
        :param subnets: The IDs of the subnets in the Neptune VPC
        :param securityGroupIds: The VPC security group IDs.
        :param volumeEncryptionKMSKey: The AWS Key Management Service (AWS KMS) key that SageMaker uses to encrypt data on the storage volume attached to the ML compute instances that run the processing job
        :param s3OutputEncryptionKMSKey: The AWS Key Management Service (AWS KMS) key that SageMaker uses to encrypt the output of the training job.
        :param customModelTransformParameters: The configuration for doing transform for a custom model. This is a JSON object with the following fields sourceDirectory, sourceS3DirectoryPath, trainingEntryPointScript, transformEntryPointScript
        :param params: a JSON dict for model transform parameters to use instead of the kwargs above to simplify parameter management
        :param wait: Whether to run job synchronously and wait for job to complete. default False

        :return: a JSON dict with the job response
        """
        dataProcessingJobId = dataProcessingJobId or params.get('dataProcessingJobId', None)
        mlModelTrainingJobId = mlModelTrainingJobId or params.get('mlModelTrainingJobId', None)
        trainingJobName = trainingJobName or params.get('trainingJobName', None)

        assert (dataProcessingJobId is not None and mlModelTrainingJobId is not None) or trainingJobName is not None, \
            "Both dataProcessingJobId and mlModelTrainingJobId are required or trainingJobName is required"

        modelTransformOutputS3Location = modelTransformOutputS3Location or params.get('modelTransformOutputS3Location',
                                                                                      None)
        assert modelTransformOutputS3Location is not None, "modelTransformOutputS3Location is required"

        if customModelTransformParameters is not None:
            customModelTransformParameters = dict(customModelTransformParameters)
            assert 'sourceS3DirectoryPath' in customModelTrainingParameters, "sourceS3DirectoryPath must be provided in customModelTrainingParameters"
            if 'sourceDirectory' in customModelTrainingParameters:
                source_directory = customModelTrainingParameters.pop('sourceDirectory')
                s3_utils.upload(source_directory, customModelTrainingParameters['sourceS3DirectoryPath'])

        if not params:
            params = {}
            if id:
                params['id'] = id
            if baseProcessingInstanceType:
                params['baseProcessingInstanceType'] = baseProcessingInstanceType
            if baseProcessingInstanceVolumeSizeInGB:
                params['trainingInstanceVolumeSizeInGB'] = trainingInstanceVolumeSizeInGB
            if customModelTransformParameters:
                params['customModelTransformParameters'] = customModelTransformParameters

            self._add_security_params(params, sagemakerIamRoleArn, neptuneIamRoleArn, subnets, securityGroupIds,
                                      volumeEncryptionKMSKey, s3OutputEncryptionKMSKey)

        model_transform_job_response = self.client.modeltransform_create(modelTransformOutputS3Location,
                                                                         dataProcessingJobId,
                                                                         mlModelTrainingJobId,
                                                                         trainingJobName, **params)
        model_transform_job_response.raise_for_status()
        model_transform_job = model_transform_job_response.json()
        if wait:
            neptuneIamRoleArn = params.get("neptuneIamRoleArn", neptuneIamRoleArn)
            model_transform_status = self.describe_model_transform_job(model_transform_job["id"],
                                                                       neptuneIamRoleArn=neptuneIamRoleArn)
            sagmaker_utils.wait_for_job(model_transform_status["processingJob"]["name"])

        return model_transform_job

    def describe_model_transform_job(self, id, neptuneIamRoleArn=None, verbose=False):
        """
        :param id: identifier of the model transform job
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker
        :param verbose: Whether to include granular job details from SageMaker Processing job

        :return: a JSON dict with the job response
        """
        neptuneIamRoleArn = neptuneIamRoleArn or self.neptune_iam_role_arn
        if neptuneIamRoleArn is not None:
            model_transform_status = self.client.modeltransform_job_status(id, neptune_iam_role_arn=neptuneIamRoleArn)
        else:
            model_transform_status = self.client.modeltransform_job_status(id)

        model_transform_status.raise_for_status()
        model_transform_job = model_transform_status.json()

        if verbose:
            model_transform_job["processingJobDetails"] = sagemaker_utils.get_processing_job_details(
                model_transform_job["processingJob"]["name"])

        return model_transform_job

    def list_model_transform_jobs(self, maxItems=10, neptuneIamRoleArn=None):
        """
        List model transform jobs associated with the current DB cluster config

        :param maxItems: number of Items to return at once
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker

        :return: a JSON list with the job response
        """
        neptuneIamRoleArn = neptuneIamRoleArn or self.neptune_iam_role_arn
        if neptuneIamRoleArn is not None:
            list_result = self.client.modeltransform_list(max_items=maxItems, neptune_iam_role_arn=neptuneIamRoleArn)
        else:
            list_result = self.client.modeltransform_list(max_items=maxItems)
        list_result.raise_for_status()
        return list_result.json()

    def stop_model_transform_job(self, id, clean=False, neptuneIamRoleArn=None):
        """
        Stop an existing model transform job.

        :param id: identifier of the model transform job
        :param clean: whether to delete any S3 artifacts created by the stopped job
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker
        """
        neptuneIamRoleArn = neptuneIamRoleArn or self.neptune_iam_role_arn
        if neptuneIamRoleArn is not None:
            stop_result = self.client.modeltransform_stop(id, clean=clean, neptune_iam_role_arn=neptuneIamRoleArn)
        else:
            stop_result = self.client.modeltransform_stop(id, clean=clean)
        stop_result.raise_for_status()

    def create_endpoint(self,
                        id=None,
                        mlModelTrainingJobId=None,
                        mlModelTransformJobId=None,
                        update=False,
                        neptuneIamRoleArn=None,
                        instanceType=None,
                        instanceCount=None,
                        volumeEncryptionKMSKey=None,
                        params={},
                        wait=False):
        """
        Create a new Neptune ML inference endpoint. See https://docs.aws.amazon.com/neptune/latest/userguide/machine-learning-api-endpoints.html#machine-learning-api-endpoints-create-job for details
        If wait is set to True, the call is synchronous and waits for endpoint creation to complete

        :param id: A unique identifier for the new endpoint. if None a uuid based id is autogenerated
        :param mlModelTrainingJobId: The job Id of the completed model-training job that has created the model that the inference endpoint host.
        :param mlModelTransformJobId: The job Id of the completed model-transform job that has created the model that the inference endpoint host
        :param update: If present, this parameter indicates that this is an update request
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker. Note: This must be listed in your DB cluster parameter group or an error will occur.
        :param instanceType: The type of ML instance used for hosting the endpoint
        :param instanceCount: The number of instances to deploy to an endpoint for prediction.
        :param volumeEncryptionKMSKey: The AWS Key Management Service (AWS KMS) key that SageMaker uses to encrypt data on the storage volume attached to the ML compute instance(s) that run the endpoints.
        :param params: a JSON dict for model transform parameters to use instead of the kwargs above to simplify parameter management
        :param wait: Whether to run job synchronously and wait for job to complete. default False

        :return: a JSON list with the create response
        """
        mlModelTrainingJobId = mlModelTrainingJobId or params.get('mlModelTrainingJobId', None)
        mlModelTransformJobId = mlModelTransformJobId or params.get('mlModelTransformJobId', None)

        assert (mlModelTrainingJobId is not None or mlModelTransformJobId is not None), \
            "Either mlModelTrainingJobId or mlModelTransformJobId is required"

        if not params:
            params = {}
            if id:
                params['id'] = id
            if update:
                params['update'] = update
            if modelName:
                params['modelName'] = modelName
            if instanceType:
                params['instanceType'] = instanceType
            if instanceCount:
                params['instanceCount'] = instanceCount
            if neptuneIamRoleArn:
                params['neptuneIamRoleArn'] = neptuneIamRoleArn
            if volumeEncryptionKMSKey:
                params['volumeEncryptionKMSKey'] = volumeEncryptionKMSKey

        endpoint_response = self.client.endpoints_create(mlModelTrainingJobId, mlModelTransformJobId, **params)
        endpoint_response.raise_for_status()
        endpoint = endpoint_response.json()
        if wait:
            neptuneIamRoleArn = params.get("neptuneIamRoleArn", neptuneIamRoleArn)
            endpoint_status = self.describe_endpoint(endpoint["id"], neptuneIamRoleArn=neptuneIamRoleArn)
            sagemaker_utils.wait_for_endpoint(endpoint_status["endpoint"]["name"])

        return endpoint

    def update_endpoint(self,
                        id=None,
                        mlModelTrainingJobId=None,
                        mlModelTransformJobId=None,
                        modelName=None,
                        instanceType=None,
                        instanceCount=None,
                        volumeEncryptionKMSKey=None,
                        params={},
                        wait=False):
        """
        Update an existing endpoint. parameters are described in create_endpoint
        """
        return self.create_endpoint(id=id,
                                    mlModelTrainingJobId=mlModelTrainingJobId,
                                    mlModelTransformJobId=mlModelTransformJobId,
                                    modelName=modelName,
                                    instanceType=instanceType,
                                    instanceCount=instanceCount,
                                    volumeEncryptionKMSKey=volumeEncryptionKMSKey,
                                    params=params,
                                    wait=wait,
                                    update=True)

    def describe_endpoint(self, id, neptuneIamRoleArn=None, verbose=False):
        """

        :param id: identifier of the endpoint
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker
        :param verbose: Whether to include granular job details from SageMaker Processing job

        :return: a JSON dict with the endpoint response
        """
        neptuneIamRoleArn = neptuneIamRoleArn or self.neptune_iam_role_arn
        if neptuneIamRoleArn is not None:
            endpoint_status = self.client.endpoints_status(id, neptune_iam_role_arn=neptuneIamRoleArn)
        else:
            endpoint_status = self.client.endpoints_status(id)

        endpoint_status.raise_for_status()
        endpoint = endpoint_status.json()

        return endpoint

    def list_endpoints(self, maxItems=10, neptuneIamRoleArn=None):
        """
        List model inference endpoints associated with the current DB cluster config

        :param maxItems: number of Items to return at once
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker

        :return: a JSON list with the endpoint response
        """

        neptuneIamRoleArn = neptuneIamRoleArn or self.neptune_iam_role_arn
        if neptuneIamRoleArn is not None:
            list_result = self.client.endpoints(max_items=maxItems, neptune_iam_role_arn=neptuneIamRoleArn)
        else:
            list_result = self.client.endpoints(max_items=maxItems)
        list_result.raise_for_status()
        return list_result.json()

    def delete_endpoint(self, id, neptuneIamRoleArn=None):
        """
        Delete an existing model inference endpoint.

        :param id: identifier of the endpoint
        :param neptuneIamRoleArn: The ARN of an IAM role for calling SageMaker
        """
        neptuneIamRoleArn = neptuneIamRoleArn or self.neptune_iam_role_arn
        if neptuneIamRoleArn is not None:
            delete_result = self.client.endpoints_delete(id, neptune_iam_role_arn=neptuneIamRoleArn)
        else:
            delete_result = self.client.endpoints_delete(id)
        delete_result.raise_for_status()

    def get_node_index_mapping(self, data_processing_job_id=None, vertex_label=None, download_location='./model-artifacts'):
        """
        Get the mapping object for Neptune vertex ids to the DGL Graph node index

        :param data_processing_job_id: The job id of a completed data-processing job that created the DGL Graph
        :param vertex_label: The vertex label you want a mapping for. if None, then all mappings for all vertex labels are returned
        :param download_location: Local path to download the mapping object

        :return: dict[node_type : dict[node_id: node_index]] For each node type/vertex label a mapping of ids to dgl node indices
        """
        assert data_processing_job_id is not None, \
            "You must provide either a data processing job id to obtain node to index mappings"

        job_details = self.describe_data_processing_job(data_processing_job_id)
        job_s3_output = job_details["processingJob"]["outputLocation"]

        # get mappings
        download_location = os.path.join(download_location, data_processing_job_id)
        s3_utils.download(os.path.join(job_s3_output, DEFAULT_PROCESSING_MAPPING_INFO), download_location)

        with open(os.path.join(download_location, DEFAULT_PROCESSING_MAPPING_INFO), "rb") as f:
            info = pickle.load(f)
            mapping = info[DEFAULT_PROCESSING_MAPPING_KEY]
            embedding_index_mapping = None
            if vertex_label is not None:
                if vertex_label in mapping:
                    mapping = mapping[vertex_label]
                else:
                    print("Mapping for vertex label: {} not found.".format(vertex_label))
                    print("valid vertex labels which have vertices mapped to embeddings: {} ".format(
                        list(mapping.keys())))
                    print("Returning mapping for all valid vertex labels")
        return mapping, embedding_index_mapping

    def get_embeddings(self, model_training_job_id, download_location='./model-artifacts', kms_key=None):
        """
        Get the node embeddings generated by a Neptune ML training job

        :param model_training_job_id: The job id of a completed model-training job that created the embeddings
        :param download_location: Local path to download the embedding file
        :param kms_key: AWS KMS key that SageMaker uses to encrypt/decrypt the embeddings in S3

        :return: numpy.NDArray. A dense embedding ndarray
        """
        assert model_training_job_id is not None, "model_training_job_id is required"
        training_job_s3_output = self.describe_model_training_job(model_training_job_id)["processingJob"][
            "outputLocation"]
        download_directory = os.path.join(download_location, model_training_job_id, "embeddings")

        s3_utils.download(os.path.join(training_job_s3_output, "embeddings"), download_directory, kms_key=kms_key)
        entity_emb = np.load(os.path.join(download_directory, DEFAULT_EMBEDDING_FILE_NAME))

        return entity_emb

    def get_predictions(self, model_training_job_id, download_location='./model-artifacts', class_preds=False,
                        kms_key=None):
        """
        Get the node predictions generated by a Neptune ML training job

        :param model_training_job_id: The job id of a completed model-training job that created the predictions
        :param download_location: Local path to download the embedding file
        :param class_preds: For classification predictions whether to apply argmax to return predicted class label
        :param kms_key: AWS KMS key that SageMaker uses to encrypt/decrypt the embeddings in S3

        :return: numpy.NDArray. An ndarray for the predictions
        """
        assert model_training_job_id is not None, "model_training_job_id is required"
        training_job_s3_output = self.describe_model_training_job(model_training_job_id)["processingJob"][
            "outputLocation"]
        download_directory = os.path.join(download_location, model_training_job_id, "predictions")

        s3_utils.download(os.path.join(training_job_s3_output, "predictions"), download_directory, kms_key)
        preds = np.load(os.path.join(download_directory, DEFAULT_PREDICTION_FILE_NAME))['infer_scores']

        if class_preds:
            return preds.argmax(axis=1)

        return preds

    def get_model_performance_metrics(self, model_training_job_id):
        """
        Return validation and test metrics for a Neptune ML model traning job

        :param model_training_job_id: The job id of a completed model-training job
        :return: JSON dict containing model performance metrics
        """
        assert model_training_job_id is not None, "model_training_job_id is required"
        training_job_s3_output = self.describe_model_training_job(model_training_job_id)["processingJob"][
            "outputLocation"]
        metrics = {}
        metrics['validation'] = s3_utils.get_config(training_job_s3_output, DEFAULT_EVAL_METRIC_FILE)
        try:
            metrics['test'] = s3_utils.get_config(training_job_s3_output, DEFAULT_TEST_METRIC_FILE)
        except:
            print("Test metrics not present, Returning only validation metrics")
        return metrics

    def get_endpoint_instance_recommendation(self, model_training_job_id=None, model_transform_job_id=None):
        """
        Get configuration for recommended inference endpoint instance from the file that is created by a Neptune ML model training job or a Neptune ML model transform job

        :param model_training_job_id: The job id of a completed model-training job
        :param model_transform_job_id: The job id of a completed model-transform job
        :return: JSON dict containing instance recommendations
        """
        assert model_training_job_id is not None or model_transform_job_id is not None, \
            "You must provide either a model training job id or a model transform job id to get endpoint instance recommendation"

        job_id = model_training_job_id if model_training_job_id is not None else model_transform_job_id
        job_details = self.describe_model_training_job(
            job_id) if job_id == model_training_job_id else self.model_transform_job_id(job_id)
        job_s3_output = job_details["processingJob"]["outputLocation"]
        return s3_utils.get_config(job_s3_output, DEFAULT_INFER_INSTANCE_REC)

    def invoke_endpoint(self, task="link_predict", endpoint_id=None, endpoint_name=None, headNodeId=None,
                        tailNodeId=None,
                        headNodeType=None, tailNodeType=None, edgeType=None, property=None, topk=1):
        """
        invoke a Neptune ML endpoint directly in python for prediction on a single item i.e node or edge

        :param task: The task that the model on the endpoint was trained for options ["node_property_prediction", "edge_property_prediction", "link_prediction"]
        :param endpoint_id: The id of a created endpoint
        :param endpoint_name: The name of a SageMaker endpoint
        :param headNodeId: The main node id for prediction for node property prediction and outgoing link prediction
        :param tailNodeId: The main node id for incoming link prediction. Also used for edge property prediction with headNodeId to define the edge for prediction
        :param headNodeType: The node type or vertex label of headNodeId
        :param tailNodeType: The node type or vertex label of tailNodeId
        :param edgeType: The edge type or label of (headNodeId, tailNodeId) for edge property prediction tasks
        :param property: The property to predict for the node or edge for property prediction types
        :param topk: How many of the top results to return

        :return: JSON response from the inference endpoint
        """
        if endpoint_name is None:
            assert endpoint_id is not None, "endpoint id is a required argument if endpoint name is missing"
            endpoint_name = self.describe_endpoint(endpoint_id)["endpoint"]["name"]
        if task in ["link_predict", "link_prediction"]:
            task = "link_predict"
            assert edgeType is not None, "edgeType is required for link prediction"
            if headNodeType is not None and tailNodeId is not None and edgeType is not None:
                assert headNodeId is None, "headNodeId should not be be passed for link prediction when tailNodeId is passed"
            elif tailNodeType is not None and headNodeId is not None and edgeType is not None:
                assert tailNodeId is None, "tailNodeId should not be be passed for link prediction when headNodeId is passed"
            else:
                print(
                    "For link prediction you must pass the pair (headNodeType, tailNodeId) or (tailNodeType , headNodeid)")
                raise RuntimeError
        elif task in ["edge_predict", "edge_prediction", "edge_property_prediction", "edge_classification",
                      "edge_regression"]:
            task = "edge_predict"
            assert headNodeId is not None and tailNodeId is not None, "headNodeId and tailNodeId is required for edge prediction"
            assert edgeType is not None, "edgeType should be passed for edge prediction"
            assert property is not None, "property should be passed in for edge prediction"

        elif task in ["node_predict", "node_prediction", "node_property_prediction", "node_classification",
                      "node_regression"]:
            task = "edge_predict"
            assert tailNodeId is None, "tailNodeId should not be be passed for node prediction."
            assert headNodeId is not None, "headNodeId should be passed in for node prediction"
            assert property is not None, "property should be passed in for node prediction"
        elif task in ["knn_predict", "knn_prediction"]:
            task = "knn_predict"
            assert tailNodeId is None, "tailNodeId should not be be passed for knn prediction."
            assert headNodeId is not None, "headNodeId should be passed in for knn prediction"
            assert tailNodeType is not None, "tailNodeType is required for knn prediction"
        else:
            print("Unsupported task type : {}".format(task))
            raise RuntimeError

        input_data = {}
        input_data["version"] = "v1"
        if task == "link_predict":
            if headNodeId is not None:
                input_data["mode"] = "predict_tail"
                input_data["data"] = {
                    "globalParameters": {
                        "topk": topk,
                        "edgeType": edgeType,
                        "tailNodeType": tailNodeType,
                        "exclude_flag": 'mask'
                    },
                    "edges": [
                        {
                            "headNodeId": headNodeId
                        }
                    ]
                }
            else:
                input_data["mode"] = "predict_head"
                input_data["data"] = {
                    "globalParameters": {
                        "topk": topk,
                        "edgeType": edgeType,
                        "headNodeType": headNodeType,
                        "exclude_flag": 'mask'
                    },
                    "edges": [
                        {
                            "tailNodeId": tailNodeId
                        }
                    ]
                }
        elif task == "edge_predict":
            input_data["data"] = {
                "globalParameters": {
                    "topk": topk,
                    "edgeType": edgeType,
                    "property": property
                },
                "edges": [
                    {
                        "headNodeId": headNodeId,
                        "tailNodeId": tailNodeId
                    }
                ]
            }
        elif task == "node_predict":
            input_data["data"] = {
                "globalParameters": {
                    "topk": topk,
                    "property": property
                },
                "nodes": [
                    {
                        "nodeId": headNodeId
                    }
                ]
            }
        elif task == "knn_predict":
            input_data["mode"] = "embed_knn"
            input_data["data"] = {
                "globalParameters": {
                    "topk": topk,
                    "tailNodeType": tailNodeType,
                    "exclude_flag": 'mask'
                },
                "nodes": [
                    {
                        "nodeId": headNodeId
                    }
                ]
            }
        else:
            print("Unsupported task type : {}".format(task))
            raise RuntimeError

        return sagemaker_utils.invoke_endpoint(endpoint_name, input_data)


import numpy as np
from collections.abc import Mapping, Sequence

import dgl
from dgl import backend as F, convert, transform
from dgl.sampling.randomwalks import random_walk
from dgl.sampling.neighbor import select_topk
from dgl.base import EID
from dgl import utils

class RandomWalkNeighborSampler(object):
    """PinSage-like neighbor sampler extended to any heterogeneous graphs.
    Given a heterogeneous graph and a list of nodes, this callable will generate a subgraph
    graph where the neighbors of each given node are the most commonly visited nodes of at the end of a metapath
    by performing multiple random walks starting from that given node.  Each random walk consists
    of , with a probability of termination after each traversal.

    This is a generalization of PinSAGE sampler which only works on bidirectional bipartite
    graphs.
    Parameters
    ----------
    G : DGLGraph
        The graph.  It must be on CPU.
    num_traversals : int
        The maximum number of metapath-based traversals for a single random walk.
        Usually considered a hyperparameter.
    termination_prob : float
        Termination probability after each metapath-based traversal.
        Usually considered a hyperparameter.
    num_random_walks : int
        Number of random walks to try for each given node.
        Usually considered a hyperparameter.
    num_neighbors : int
        Number of neighbors (or most commonly visited nodes) to select for each given node.
    metapath : list[str] or list[tuple[str, str, str]], optional
        The metapath.
        If not given, DGL assumes that the graph is homogeneous and the metapath consists
        of one step over the single edge type.
    weight_column : str, default "weights"
        The name of the edge feature to be stored on the returned graph with the number of
        visits.
    Examples
    --------
    See examples in :any:`PinSAGESampler`.
    """
    def __init__(self, G, num_traversals, termination_prob,
                 num_random_walks, num_neighbors, metapath=None, weight_column='weights'):
        assert G.device == F.cpu(), "Graph must be on CPU."
        self.G = G
        self.weight_column = weight_column
        self.num_random_walks = num_random_walks
        self.num_neighbors = num_neighbors
        self.num_traversals = num_traversals

        if metapath is None:
            if len(G.ntypes) > 1 or len(G.etypes) > 1:
                raise ValueError('Metapath must be specified if the graph is homogeneous.')
            metapath = [G.canonical_etypes[0]]
        self.start_ntype = G.to_canonical_etype(metapath[0])[0]
        self.end_ntype = G.to_canonical_etype(metapath[-1])[-1]

        self.metapath_hops = len(metapath)
        self.metapath = metapath
        self.full_metapath = metapath * num_traversals
        restart_prob = np.zeros(self.metapath_hops * num_traversals)
        restart_prob[self.metapath_hops::self.metapath_hops] = termination_prob
        self.restart_prob = F.zerocopy_from_numpy(restart_prob)

    # pylint: disable=no-member
    def __call__(self, seed_nodes):
        """
        Parameters
        ----------
        seed_nodes : Tensor
            A tensor of given node IDs of node type ``ntype`` to generate neighbors from.  The
            node type ``ntype`` is the beginning and ending node type of the given metapath.
            It must be on CPU and have the same dtype as the ID type of the graph.
        Returns
        -------
        g : DGLGraph
            A homogeneous graph constructed by selecting neighbors for each given node according
            to the algorithm above.  The returned graph is on CPU.
        """
        seed_nodes = utils.prepare_tensor(self.G, seed_nodes, 'seed_nodes')

        seed_nodes = F.repeat(seed_nodes, self.num_random_walks, 0)
        paths, _ = random_walk(
            self.G, seed_nodes, metapath=self.full_metapath, restart_prob=self.restart_prob)
        src = F.reshape(paths[:, self.metapath_hops::self.metapath_hops], (-1,))
        dst = F.repeat(paths[:, 0], self.num_traversals, 0)

        src_mask = (src != -1)
        src = F.boolean_mask(src, src_mask)
        dst = F.boolean_mask(dst, src_mask)

        # count the number of visits and pick the K-most frequent neighbors for each node
        if self.start_ntype == self.end_ntype:
            self.ntype = self.start_ntype
            neighbor_graph = convert.heterograph(
                {(self.ntype, '_E', self.ntype): (src, dst)},
                {self.ntype: self.G.number_of_nodes(self.ntype)}
            )
        else:
            neighbor_graph = convert.heterograph(
                {(self.end_ntype, '_E', self.start_ntype): (src, dst)},
                {self.start_ntype: self.G.number_of_nodes(self.start_ntype),
                 self.end_ntype: self.G.number_of_nodes(self.end_ntype)}
            )

        neighbor_graph = transform.to_simple(neighbor_graph, return_counts=self.weight_column)
        counts = neighbor_graph.edata[self.weight_column]
        neighbor_graph = select_topk(neighbor_graph, self.num_neighbors, self.weight_column)
        selected_counts = F.gather_row(counts, neighbor_graph.edata[EID])
        neighbor_graph.edata[self.weight_column] = selected_counts

        return neighbor_graph

class MetapathListSampler(object):
    r"""
    Sample subgraphs for a list of metapaths by perform random walks for each metapath.

    Parameters
    ----------
    g : DGLGraph
        heterogeneous dgl graph
    metapath_list : list[list[str or tuple of str]]
        list of metapaths. each metapath is specified as a list of edge_types or canonical_edge_types.
    num_neighbors: int or list, optional
        Number of neighbors (or most commonly visited nodes) to select for each given node. (Default: 1)
    num_random_walks : int, optional
        Number of random walks to try for each given node. (Default: None)
        This is set ot num_neighbours by default. If num_random_walks > num_neighbors and the
        random walks find more than num_neighbours neighbours for a particular node the fanout
        for that node is still capped at num_neighbors
    dropout_rate : float, optional
        Dropout rate (Default: 0.0)
    add_self_loop: bool
    """
    def __init__(self, g, metapath_list, num_neighbors=1, num_random_walks=None):
        metapath_list = [[g.to_canonical_etype(etype) for etype in metapath] for metapath in metapath_list]
        if isinstance(num_neighbors, Sequence):
            for metapath in metapath_list:
                assert metapath[0][0] == metapath[-1][-1], \
                    "For multilayer sampling all metapaths must start and end at the same node type"
        if not isinstance(num_neighbors, Sequence):
            assert type(num_neighbors) == int
            num_neighbors = [num_neighbors]
        if not num_random_walks:
            num_random_walks = num_neighbors
        else:
            assert type(num_random_walks) == int
            num_random_walks = [num_random_walks]*len(num_neighbors)
        self.sampler_list = []
        self.metapaths_src_ntype = []
        for metapath in metapath_list:
            metapath_layer_sampler_list = []
            for neighbour_num, rw_num in zip(num_neighbors, num_random_walks):
                metapath_layer_sampler_list.append(RandomWalkNeighborSampler(G=g,
                                                                           num_traversals=1,
                                                                           termination_prob=0,
                                                                           num_random_walks=rw_num,
                                                                           num_neighbors=neighbour_num,
                                                                           metapath=metapath))
            self.sampler_list.append(metapath_layer_sampler_list)
            self.metapaths_src_ntype.append(g.to_canonical_etype(metapath[0])[0])
        self.num_target_ntypes = len(set(self.metapaths_src_ntype))

    def sample_blocks(self, seeds):
        r"""
        Perform one round of sampling for seed nodes
        Parameters
        ----------
        seeds : th.Tensor or dict[ntype: th.Tensor]
            seed nodes of type 'ntype' where ``ntype`` is the ending node type of the given metapath

        Returns
        -------
        (th.Tensor, list[DGLGraph])
            the input seeds which will be the target nodes for each metapath subgraph

            A list of MFGs of per sampled metapath sub-graphs necessary for computing the representation
            of the seed nodes for each metapath sub-graph
        """
        if isinstance(seeds, Mapping):
            assert all([ntype in seeds for ntype in self.metapaths_src_ntype]), \
                "All node types in {} should be present in seeds".format(set(self.metapaths_src_ntype))
        else:
            assert self.num_target_ntypes == 1, \
                "seeds should be dictionary with seeds per node type if the number of target node types is more than 1"
        blocks = []
        for i, metapath_sampler in enumerate(self.sampler_list):
            target_ntype = self.metapaths_src_ntype[i]
            ntype_input_seeds = seeds[target_ntype] if isinstance(seeds, Mapping) else seeds
            if len(metapath_sampler) == 1: ## only one layer of message passing in the computational graph
                frontier = metapath_sampler[0](ntype_input_seeds)
                # add self loop if homogeneous
                if len(frontier.ntypes) == 1:
                    block = dgl.to_block(frontier, ntype_input_seeds)
                else:
                    block = dgl.to_block(frontier, {target_ntype: ntype_input_seeds})
                blocks.append(block)
            else: ## multiple layers of message passing in the computational graph
                layer_blocks = []
                for sampler in metapath_sampler:
                    frontier = sampler(ntype_input_seeds)
                    assert len(frontier.ntypes) == 1, "For multiple layers of message passing"
                    block = dgl.to_block(frontier, ntype_input_seeds)
                    ntype_input_seeds = block.srcdata[dgl.NID]
                    layer_blocks.insert(0, block)
                blocks.append(layer_blocks)
        return blocks

    def sample(self, seeds):
        r"""
        Sample subgraphs for seed nodes

        Parameters
        ----------
        seeds : th.Tensor or dict[ntype: th.Tensor]
            seed nodes of type 'ntype' where ``ntype`` is the ending node type of the given metapath

        Returns
        -------
        list[DGLGraph]

            A list of Subgraph of per sampled metapath
        """
        if isinstance(seeds, Mapping):
            assert all([ntype in seeds for ntype in self.metapaths_src_ntype]), \
                "All node types in {} should be present in seeds".format(set(self.metapaths_src_ntype))
        else:
            assert self.num_target_ntypes == 1, \
                "seeds should be dictionary with seeds per node type if the number of target node types is more than 1"
        frontiers = []
        for i, sampler in enumerate(self.sampler_list):
            target_ntype = self.metapaths_src_ntype[i]
            ntype_input_seeds = seeds[target_ntype] if isinstance(seeds, Mapping) else seeds
            frontiers.append(sampler[0](ntype_input_seeds))

        return frontiers


if __name__ == "__main__":
    from ogb.nodeproppred import DglNodePropPredDataset
    dataset = DglNodePropPredDataset(name='ogbn-mag')
    graph, label = dataset[0]
    train_nids = dataset.get_idx_split()["train"]
    metapath_list = [['cites']]
    sampler = MetapathListSampler(graph, metapath_list, [5, 5])

    # use sampler directly
    random_train_nids = {'paper': np.random.choice(train_nids['paper'], 10)}
    # metapath_subgraphs = sampler.sample(random_train_nids)
    # for metapath, metapath_subg in zip(metapath_list, metapath_subgraphs):
    #     print(metapath)
    #     print(metapath_subg)

    metapath_block_subgraphs = sampler.sample_blocks(random_train_nids)
    for metapath, metapath_block_subg in zip(metapath_list, metapath_block_subgraphs):
        print(metapath)
        print(metapath_block_subg)



import torch as th
from torch.utils.data import DataLoader

import dgl
from dgl import utils

from collections.abc import Mapping, Sequence

class MultiSubgraphNodeDataLoader:
    r"""
    Pytorch Dataloader to be used with a Graph Sampler that returns multiple
    distinct subgraph views

    Parameters
    ----------
    g : DGLGraph
        heterogeneous dgl graph
    nids: th.Tensor or dict[str: th.Tensor]
        the dataset node ids to sample from
    sampler: int
        the sampler to generate multiple subgraph views for each batch
    """

    def __init__(self, g, nids, sampler, device=None, **kwargs):
        if device is None:
            # default to the same device the graph is on
            device = th.device(g.device)
        self.collator = MultiSubgraphNodeCollator(g, nids, sampler, device)
        self.dataloader = DataLoader(
            self.collator.dataset,
            collate_fn=self.collator.collate,
            **kwargs)

    def __iter__(self):
        """Return the iterator of the data loader."""
        return iter(self.dataloader)

    def __len__(self):
        """Return the number of batches of the data loader."""
        return len(self.dataloader)

class MultiSubgraphNodeCollator:
    r"""
    NodeCollator for multi-subgraph graph sampler

    At each minibatch multiple subgraphs are returned by the sampler
    and each subgraph is a different view of the original graph
    around the current minibatch seeds

    Parameters
    ----------
    g : DGLGraph
        heterogeneous dgl graph
    nids: th.Tensor or dict[str: th.Tensor]
        the dataset node ids to sample from
    sampler: int
        the sampling method to use to generate multiple subgraph views
    """
    def __init__(self, g, nids, sampler, device=None):
        self.g = g
        self.sampler = sampler
        self.device = device
        if isinstance(nids, Mapping):
            self._dataset = utils.FlattenedDict(nids)
        else:
            self._dataset = utils.prepare_tensor(g, nids, 'nids')

    @property
    def dataset(self):
        return self._dataset

    def collate(self, items):
        r"""
        Sample subgraphs and collate for mini-batch training

        Parameters
        ----------
        items : list[int] or list[tuple[str, int]]
            starting seed nodes

        Returns
        -------
         tuple(list[th.Tensor] or list[dict[str: th.Tensor]], th.Tensor or dict[str: th.Tensor], list[DGLGraph])
            i.e (input_node_list, output_nodes, subgraph_list)
            a tuple with list of input node ids for each subgraph,
            the target node ids and
            a list of MSGs for each subgraph
        """
        if isinstance(items[0], tuple):
            # returns a list of pairs: group them by node types into a dict
            seed_nodes = {ntype: th.stack(seeds).reshape(-1) for ntype, seeds in utils.group_as_dict(items).items()}
        else:
            seed_nodes = utils.prepare_tensor(self.g, items, 'items')
        blocks = self.sampler.sample_blocks(seed_nodes)
        output_nodes = []
        input_nodes = []
        for block in blocks:
            if isinstance(block, list): # check if this is a multi-layer MFG sequence
                input_nodes.append(block[0].srcdata[dgl.NID])
                output_nodes.append(block[-1].dstdata[dgl.NID])
            else:
                input_nodes.append(block.srcdata[dgl.NID])
                output_nodes.append(block.dstdata[dgl.NID])
        if self.device is not None:
            input_nodes = [{k: v.to(self.device) for k,v in input_node.items()} if isinstance(input_node, Mapping)
                           else input_node.to(self.device) for input_node in input_nodes]
            output_nodes = [{k: v.to(self.device) for k,v in output_node.items()} if isinstance(output_node, Mapping)
                           else output_node.to(self.device) for output_node in output_nodes]
            blocks = [[b.to(self.device) for b in block] if isinstance(block, Sequence)
                      else block.to(self.device) for block in blocks]
        return input_nodes, output_nodes, blocks